#Cody Tramp
#2013-06-26

'''MetaCom: Automated Data Processing and Analysis of Metagenomic Community Sequences.'''

#imports from main python libraries
from __future__ import division
import os
import shutil
import string
import math
import urllib
import sys
import csv
import re
import time
import datetime
import wx
import threading
import StringIO
import traceback
import random
import numpy as np
import sqlite3 as dbapi
import Bio.Entrez as entrez
from wx.lib.pubsub import setupkwargs
from wx.lib.pubsub import pub
from Queue import Queue
from Bio.Blast import NCBIWWW
from Bio.Blast import NCBIXML
from multiprocessing import Pool, freeze_support, cpu_count

##Program Information
PROGRAM_NAME = "MetaCom"
PROGRAM_VERSION = "1.1.0"
DOWNLOAD_LINK = "http://digitalcommons.usu.edu/all_datasets/1/"
CONTACT_EMAIL = "c.a.tramp@aggiemail.usu.edu"

##Images Locations
IMAGE_PATH_INFO = "InfoSymbol.png"
IMAGE_PATH_WARNING = "WarningSign.png"
IMAGE_HEIGHT_PX = 20
IMAGE_WIDTH_PX = 20
MAIN_WINDOW_WIDTH = 950
MAIN_WINDOW_HEIGHT = 505

##Button Definitions
ID_BLAST_START = wx.NewId()
ID_GET_SEQUENCE_TAG_SUMMARY = wx.NewId()
ID_COMBINE_SQL_FILES = wx.NewId()
ID_GET_TAXONOMY = wx.NewId()
ID_GET_RECOMMENDED_HITS = wx.NewId()
ID_GET_PERCENT_IDENTITY_SUMMARY = wx.NewId()
ID_CREATE_FILE_FROM_TAG_SUBSET = wx.NewId()
ID_TAXONOMY_SUMMARY_FULL_LEVELS = wx.NewId()
ID_TAXONOMY_SUMMARY_SINGLE_LEVELS = wx.NewId()
ID_GET_RAREFACTION_CURVES = wx.NewId()
ID_GET_SHANNON_INDEX = wx.NewId()
ID_EXPORT_ALL_TABLES_TO_EXCEL = wx.NewId()

## Define notification events
EVT_DATACOLLECTRESULT_ID = wx.NewId()
EVT_ASCCOLLECTRESULT_ID = wx.NewId()
EVT_BLOCKRESULT_ID = wx.NewId()
EVT_BLASTRESULT_ID = wx.NewId()
EVT_MISSINGRESULT_ID = wx.NewId()
EVT_TAXONOMYRESULT_ID = wx.NewId()
EVT_UPDATESTATUS_ID = wx.NewId()

##Define ComboBox Index Lists
cbxlist_MaxResults = []
cbxlist_TaxonomyLevels = []
cbxlist_MaxResults = ["10", "50", "100", "250", "500", "1000", "5000", "10000", "20000"]
cbxlist_TaxonomyLevels = ["Kingdom", "Phylum", "Class", "Order", "Family", "Genus", "Species"]

def EVT_UPDATESTATUS(window, function):
    '''Defines the blast block result event.'''
    window.Connect(-1, -1, EVT_UPDATESTATUS_ID, function)
def EVT_BLOCKRESULT(window, function):
    '''Defines the blast block result event.'''
    window.Connect(-1, -1, EVT_BLOCKRESULT_ID, function)

aboutText = """MetaCom Version %(version)s

MetaCom is a program that offers a common set of analysis and
organizational tools that can be used by research groups in the analysis of
environmental samples. MetaCom is an open-source software tool that can process
next generation sequencing output, identify the species present in the sample,
organize the results of the analysis in an easy to manipulate format, and can
calculate useful metrics of data quality.

Availablity: The MetaCom program, source code, user manual, examples files,
and documentation are available at: %(download)s

Contact: %(email)s

This program is running on version %(wxpy)s of wxPython and %(python)s of Python."""

class AboutBox(wx.Dialog):
    def __init__(self):
        self.vers = {}
        self.vers["version"] = PROGRAM_VERSION
        self.vers["python"] = sys.version.split()[0]
        self.vers["wxpy"] = wx.VERSION_STRING
        self.vers["email"] = CONTACT_EMAIL
        self.vers["download"] = DOWNLOAD_LINK
        self.strText = aboutText % self.vers
    
    def GetAbout(self):
        return self.strText 

class ResultEvent(wx.PyEvent):
    '''Event to carry result data to main window and can also
    update the progress of the thread to the main info window'''
    def __init__(self, data, event_ID):
        '''Init Result Event with the supplied data and event_ID'''
        wx.PyEvent.__init__(self)
        self.SetEventType(event_ID)
        self.data = data

##Class for all the options used in the program
class Options:
    '''A Class for options related to BLAST searches and 454 data.'''
    
    #Member Variables
    #strProjectName - string name of the project for saved files
    #fna_filepaths - source FNA/FASTA/SQLITE file locations
    #strIDTag - string name/initials to rename sequences with "<IDTAG>1, ..."
    #boolGetTaxonomyDuringBLAST - boolean indicating if taxonomy search done in parallel with BLAST run
    #boolMultipleComputerAnalysis - boolean indicating if computer is one of many running same sample
    #intMultipleComputerTotalNumber - integer indicating which how many computer in processing group, used for splitting workload 
    #intMultipleComputerIndex - integer indicating which computer in group it is, used for splitting workload
    #intBLASTMinimumAlignmentLength - integer indicating minimum required length of BLAST alignment for inclusion in save
    #intBLASTEvalueCutoffExponent - integer indicating what minimum e-value 1e^-(value) to require for includion in save
    #indexBLASTMaxResultsPerSequence - int indicating index of sequence results to return [10,50,100,250,500,1000,5000,10000,20000]
    #intBLASTSimultaneousThreads - int indicating number of parallel calls to BLAST server to run (BLAST rules require max = 2, overidable)
    #boolBLASTSaveAllHits - boolean indicating whether all returned results will be included in output data set
    #boolBlastSaveUnculturedClones - boolean indicating whether all uncultured clone and unidentified species hits will be included in output data set
    #indexTaxonomyFullOutputLevel - index indicating which lower level of taxomony to go down to in a full report [Kingdom, Phylum, Class, Order, Family, Genus, Species]
    #indexTaxonomySingleOutputLevel - index indicating which single level of taxomony to output in report[Kingdom, Phylum, Class, Order, Family, Genus, Species]
    #intStartingPosition - integer indicating the position of the last sequence analyze (for faster restarting of analyses)
  
    def __init__(self, fna_filepaths, strProjectName="MetaCom Project", strEmail = "none@none.com", strIDTag="Seq",
                 boolMultipleComputerAnalysis=False, intMultipleComputerIndex = 1, intMultipleComputerTotalNumber=1,
                 intBLASTMinimumAlignmentLength=1, intBLASTEvalueCutoffExponent=6, indexBLASTMaxResultsPerSequence=3,
                 intBLASTSimultaneousThreads=2, boolBLASTSaveAllHits=False, boolBlastSaveUnculturedClones=False,
                 indexTaxonomyFullOutputLevel=0, indexTaxonomySingleOutputLevel=0, intStartingPostion=0):
        
        self.fna_filepaths = []
        for filepath in fna_filepaths:
            self.fna_filepaths.append(filepath)
            
        self.strProjectName = strProjectName
        self.strEmail = strEmail
        self.strIDTag = strIDTag
        self.boolMultipleComputerAnalysis = boolMultipleComputerAnalysis
        self.intMultipleComputerIndex = intMultipleComputerIndex        
        self.intMultipleComputerTotalNumber = intMultipleComputerTotalNumber
        self.intBLASTMinimumAlignmentLength = intBLASTMinimumAlignmentLength
        self.intBLASTEvalueCutoffExponent = intBLASTEvalueCutoffExponent
        self.strBLASTEvalue = self.getEvalueFromExponent(self.intBLASTEvalueCutoffExponent)
        self.indexBLASTMaxResultsPerSequence = indexBLASTMaxResultsPerSequence
        self.intBLASTSimultaneousThreads = intBLASTSimultaneousThreads
        self.boolBLASTSaveAllHits = boolBLASTSaveAllHits
        self.boolBlastSaveUnculturedClones = boolBlastSaveUnculturedClones
        self.indexTaxonomyFullOutputLevel = indexTaxonomyFullOutputLevel
        self.indexTaxonomySingleOutputLevel = indexTaxonomySingleOutputLevel
        self.intStartingPosition = intStartingPostion
        
    def SaveOptions(self):
        '''Saves the options on the form to the "Options.txt" file for later use.'''
        output_filename = (os.path.dirname(os.path.realpath(__file__)) + "\MetaCom_Options.txt")
        
        output_file = open(output_filename, 'wb')
        
        output_file.writelines("Project Name == " + self.strProjectName + "\n")
        output_file.writelines("Email Address == " + self.strEmail + "\n")
        output_file.writelines("Data Filepaths == " + "" + "\n")
        output_file.writelines("Initials/IDtag == " + self.strIDTag + "\n")
        output_file.writelines("Multiple Computer Analysis (True/False) == " + str(self.boolMultipleComputerAnalysis)+ "\n")
        output_file.writelines("Multiple Computer Index (this computer) == " + str(self.intMultipleComputerIndex) + "\n")
        output_file.writelines("Multiple Computer Total Number == " + str(self.intMultipleComputerTotalNumber) + "\n")
        output_file.writelines("Minimum BLAST alignment length == " + str(self.intBLASTMinimumAlignmentLength) + "\n")
        output_file.writelines("BLAST E-value Cutoff Exponent (1e^-x) == " + str(self.intBLASTEvalueCutoffExponent) + "\n")
        output_file.writelines("Max BLAST results to return per sequence == " + str(self.indexBLASTMaxResultsPerSequence) + "\n")
        output_file.writelines("Number of parallel BLAST threads to run == " + str(self.intBLASTSimultaneousThreads) + "\n")
        output_file.writelines("Save All Blast Hits(True/False) == " + str(self.boolBLASTSaveAllHits) + "\n")
        output_file.writelines("Save Uncultured/Unidentified BLAST hits (True/False) == " + str(self.boolBlastSaveUnculturedClones) + "\n")
        output_file.writelines("Taxonomy Level INDEX to Output Full Taxonomy Report == " + str(self.indexTaxonomyFullOutputLevel) + "\n")
        output_file.writelines("Taxonomy Level INDEX to Output Single Level Report == " + str(self.indexTaxonomySingleOutputLevel) + "\n")
        output_file.writelines("Starting Position For Sequence Analysis == " + str(self.intStartingPosition) + "\n") 
        
        output_file.close()    
        
    def LoadOptions(self, options_filename):
        '''Loads an option file from options_filepath. This would allow old settings to be reused
        in a given directory. Each entry's name and value is seperated by " == " in the file.'''
        
        input_file = open(options_filename, 'rU')
        option_lines = input_file.readlines()
        input_file.close()              
        
        for line in option_lines:
            splitline = line.split(" == ")
            if splitline[0] == "Project Name":
                self.strProjectName = splitline[1].strip()
            elif splitline[0] == "Email Address":
                self.strEmail = splitline[1].strip()
            elif splitline[0] == "Data Filepaths":
                self.fna_filepaths = self.GetListFromString(splitline[1])
            elif splitline[0] == "Initials/IDtag":
                self.strIDTag = splitline[1].strip()                
            elif splitline[0] == "Multiple Computer Analysis (True/False)":
                self.boolMultipleComputerAnalysis = self.GetBooleanFromString(splitline[1])              
            elif splitline[0] == "Multiple Computer Index (this computer)":
                self.intMultipleComputerIndex = int(splitline[1].strip())
            elif splitline[0] == "Multiple Computer Total Number":
                self.intMultipleComputerTotalNumber = int(splitline[1].strip())                
            elif splitline[0] == "Minimum BLAST alignment length":
                self.intBLASTMinimumAlignmentLength = int(splitline[1].strip())                
            elif splitline[0] == "BLAST E-value Cutoff Exponent (1e^-x)":
                self.intBLASTEvalueCutoffExponent = int(splitline[1].strip())                
            elif splitline[0] == "Max BLAST results to return per sequence":
                self.indexBLASTMaxResultsPerSequence = int(splitline[1].strip())                
            elif splitline[0] == "Number of parallel BLAST threads to run":
                self.intBLASTSimultaneousThreads = int(splitline[1].strip())                
            elif splitline[0] == "Save All Blast Hits(True/False)":
                self.boolBLASTSaveAllHits = self.GetBooleanFromString(splitline[1])
            elif splitline[0] == "Save Uncultured/Unidentified BLAST hits (True/False)":
                self.boolBlastSaveUnculturedClones = self.GetBooleanFromString(splitline[1])
            elif splitline[0] == "Taxonomy Level INDEX to Output Full Taxonomy Report":
                self.indexTaxonomyFullOutputLevel = int(splitline[1].strip())                            
            elif splitline[0] == "Taxonomy Level INDEX to Output Single Level Report":
                self.indexTaxonomySingleOutputLevel = int(splitline[1].strip())
            elif splitline[0] == "Starting Position For Sequence Analysis":
                self.intStartingPosition = int(splitline[1].strip())
            
            self.SaveOptions()

    def GetBooleanFromString(self, strToTest, boolDefaultValue=False):
        '''Returns a boolean type variable based on the string strToTest'''
        strToTest = strToTest.strip()
        if (strToTest == "True" or strToTest == "true"):
            return True
        elif (strToTest == "False" or strToTest == "false"):
            return False
        else:
            return boolDefaultValue
        
    def GetListFromString(self, *args):
        '''Returns a list object from a given input string. Removes unicode formatting.'''
        strToParse = ""
        for arg in args:
            try:        
                strToParse += arg.strip()
                ##Remove formating ("'", " ", "u")
                strToParse = strToParse.replace("u'","")
                strToParse = strToParse.replace('u"','')
                strToParse = strToParse.replace(" ","")
            except:
                pass
        return strToParse.split(",")
    
    def GetStringFromList(self, *args):
        '''Returns a string object from a given input list. Removes unicode formatting.'''
        strToParse = ""
        for arg in args:
            try:
                if type(arg) == list:
                    strToParse += str(arg)[1:-1]
                elif type(arg) == string:
                    strToParse += arg
                else:
                    strToParse += arg.toString()
                ##Remove formating (" ", "u")
                strToParse = strToParse.replace("u'","")
                strToParse = strToParse.replace('u"','')
                strToParse = strToParse.replace(" ","")
            except:
                pass
        return strToParse 
    
    def getEvalueFromExponent(self, exponent):
        return "1e-" + str(exponent)
        
            
##Class for BLAST results objects, customized from the BioPython Blast_record
##so that it contains only useful parameters
class BLASTResult:
    '''A Class for BLAST result objects, each object is one alignment match'''
    
    #Member Variables
    #Asc - ascension #
    #Name - contains the species name, and match name
    #Score - match score
    #Evalue - expected value
    #Indentities - length of match identities
    #Coverage - percent of input sequence that has identities
    #query_start - position in input string that match starts on
    #query_end - position in input string that match ends on
    #subject_start - position in subject string that match starts on
    #subject_end - position in subject string that match ends on
    
    def __init__(self, input_query_length, asc, name, score, evalue, identities, subject_as_string):
        '''Creates the BlastResult Object with the correct member variables'''
        self.asc = asc
        self.name = name
        self.score = score
        self.evalue = evalue
        self.identities = identities
        self.coverage = (self.identities/input_query_length) * 100
        
        #default start values to invalid values
        self.query_start = -1
        self.query_end = -1
        self.subject_start = -1
        self.subject_end = -1
        
        #rip start and endpoints from the summary string, since endpoint isn't
        #part of the Biopython output member variables
        subject_multiline_string = subject_as_string.split('\n')
        p = re.compile("(?P<start>\d+)(\s{1})(?P<sequence>.+)(\s{1})(?P<end>\d+)")
        for line in subject_multiline_string:
            if "Query:" in line:
                m = p.search(line)
                if m:
                    self.query_start = m.group("start")
                    self.query_end = m.group("end")
                else:
                    raise Exception("BlastParsingError",
                                    "Couldn't match query regex")
            elif "Sbjct:" in line:
                m = p.search(line)
                if m:
                    self.subject_start = m.group("start")
                    self.subject_end = m.group("end")
                else:
                    raise Exception("BlastParsingError",
                                    "Couldn't match subject regex") 
                
##Class for Seq454 Objects (essentially a stripped down SeqIO format, with
##count member variable added in to reduce BLAST runtimes
class Seq454:
    '''A Class for 454 Sequence Objects used for basic sequence metadata storage.'''
    
    #Member variables
    #name - name of the sequence (doesn't contain ">")
    #original_name - original name of sequence from file to allow backtracing
    #seq - DNA sequence string of the object
    #tag - 454 Sequencing tag for the object (could be a sequence or "other")
    #count - number of copies of this sequence present in sample
    #tophits - list of top BLAST result objects
    #allhits - list of all BLAST result objects above input match parameters
    #unculturedhits - list of all uncultured clone BLAST result objects

    def __init__(self, name, sequence, count=1):
        '''Initializes the Seq454 object. Sequence can be in the form of a list
        or a single string.'''
        self.name = name
        self.original_name = name
        self.count = int(count)
        
        if type(sequence) == list:
            self.seq = ''
            for part in sequence:
                part = part.lower()
                self.seq += part
        else:
            sequence = sequence.lower()
            self.seq = sequence
        self.seq = self.seq.rstrip()
        
        self.tag = "" #defaults to empty

        self.tophits = []
        self.allhits = []
        self.unculturedhits = []
            
    def FASTA(self):
        '''Returns a string containing the FASTA format sequence and name.'''
        return('>' + self.name + '\n' + self.seq + '\n')
    
    def details(self):
        '''Returns a string containg FASTA formated sequence and name. Name
        includes the value of count in parentheses at the end.'''
        return('>' + self.name + '(' + str(self.count) + ')\n' + self.seq + '\n')
    
    def changeseq(self, new_sequence):
        '''Changes the sequence member variable to new_sequence. Recalculates
        the length member variable of the object.'''
        new_sequence = new_sequence.lower()
        if type(new_sequence) == list:
            self.seq = ''
            for part in new_sequence:
                self.seq += part
        else:
            self.seq = new_sequence
            
        self.seq = self.seq.rstrip()
      
    def increment(self, amount=1):
        '''Increases the count variable by given amount (default = 1).'''
        self.count += amount
        
    def determine_tag(self):
        '''Determines which tag (if any) of those in the input list is present.
        Removes the tag sequence from the self.seq string (it wouldn't match to
        BLAST anyway) and updates self.tag with the correct tag. If no tag
        matches, tag remains set to "other" and no sequence is removed from
        self.seq. Returns the tag that it selected.'''
        
        #Filter tags - convert all to lowercase, strip whitespace, and remove 
        #any empty entries
        
        self.tag = self.seq[0:10]
        self.seq = self.seq[len(self.tag):]

        return self.tag

    ##WX Class for the Main Window Creation - Based on "Getting Started" Wiki page

class NotifyPanel(wx.Panel):
    def __init__(self, parent, id, notifyType, toolTipMessage):
        # create the panel
        wx.Panel.__init__(self, parent, id)

        self.SetToolTip(wx.ToolTip(toolTipMessage))
        
        if (notifyType == "warning"):
            warning_filepath = self.resource_path(IMAGE_PATH_WARNING)
            self.bg = wx.Bitmap(warning_filepath, wx.BITMAP_TYPE_PNG)
        else:
            info_filepath = self.resource_path(IMAGE_PATH_INFO)
            self.bg = wx.Bitmap(info_filepath, wx.BITMAP_TYPE_PNG)        

        self._width, self._height = self.bg.GetSize()
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        
    def resource_path(self, relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.abspath(".")
    
        return os.path.join(base_path, relative_path)        
        
    def OnPaint(self, evt):
        dc = wx.BufferedPaintDC(self)
        self.Draw(dc)
        
    def Draw(self, dc):
        cliWidth, cliHeight = self.GetClientSize()
        dc.Clear()
        xPos = (cliWidth - self._width)/2
        yPos = (cliHeight - self._height)/2
        dc.DrawBitmap(self.bg, xPos, yPos)        
        
class MainWindow(wx.Frame):
    def __init__(self, parent, title):
        #Check to see if options file exists in the current directory,
        #if not use the default options values
        options = Options([])
        
        if os.path.exists("MetaCom_Options.txt"):
            #Exisiting file of options found, read file
            options.LoadOptions("MetaCom_Options.txt")
        else:
            #File doesn't exist, use default options
            options = Options([])
            
              
        wx.Frame.__init__(self, parent, title=title,
                          size=(MAIN_WINDOW_WIDTH,MAIN_WINDOW_HEIGHT),
                          style=wx.DEFAULT_FRAME_STYLE ^ wx.RESIZE_BORDER ^ wx.MAXIMIZE_BOX)
        self.CenterOnScreen()
        
        self.sizerLeft = wx.BoxSizer(wx.VERTICAL)
        self.sizerLeftInfo = wx.BoxSizer(wx.VERTICAL)
        self.sizerRight = wx.BoxSizer(wx.VERTICAL)
        self.sizerRightInfo = wx.BoxSizer(wx.VERTICAL)
        
        self.FontTitle = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        self.FontTitle.SetUnderlined(True)
        self.FontTitle.Weight = wx.FONTWEIGHT_BOLD
        self.FontTitle.SetPointSize(15)
        #self.FontTitle.setWeight(wx.FONTWEIGHT_BOLD)
        
        self.FontSubTitle = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        self.FontSubTitle.SetUnderlined(True)
        self.FontSubTitle.SetPointSize(12)

        strBlastAnalysisTip = '''These are the options for reading new sequence files from .FASTA or .FNA
files, as well as conducting and saving the BLAST query of the sequence
reads.'''
        AboutProgram = AboutBox()
        self.infoMetaCom = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=AboutProgram.GetAbout())
        self.lblBlastAnalysisTitle = wx.StaticText(parent=self, id=-1, label="Project Analysis Options:")
        self.lblBlastAnalysisTitle.Font = self.FontTitle
        self.infoBlastAnalysisTitle = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strBlastAnalysisTip)        
        self.sizerBlastAnalysisTitle = self.AddItemsToHorizontalSizer(self.lblBlastAnalysisTitle, self.infoBlastAnalysisTitle, self.infoMetaCom)
        
        strProjectTip = '''This will be the name of the new database file for this project.

If this name differs from the one that was initially created after 
reading a .FASTA or .FNA file, then the results of any analysis
done after this point will be stored under this filename,
and not the original filename.'''
        self.lblProject = wx.StaticText(parent=self, id=-1, label="Project Name/Title:")
        self.tbxProject = wx.TextCtrl(parent=self, id=-1, value=options.strProjectName, size = (198,22))
        self.infoProject = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strProjectTip)
        self.sizerProject = self.AddItemsToHorizontalSizer(None, self.infoProject, self.lblProject, self.tbxProject)
        
        strEmailTip = '''This REQUIRES a valid email address in order to query the BLAST and NCBI 
databases. All correspondence from these sites in respect to potential
over-use and IP blocking will be directed to this email.

This email will not be used for any other purpose, and will only be 
saved to your local settings file, not with any result output.'''
        self.lblEmail= wx.StaticText(parent=self, id=-1, label="User's Email Address:")
        self.tbxEmail = wx.TextCtrl(parent=self, id=-1, value=options.strEmail, size = (188,22))
        self.infoEmail = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strEmailTip)
        self.sizerEmail= self.AddItemsToHorizontalSizer(None, self.infoEmail, self.lblEmail, self.tbxEmail)        

        strIDTagTip = '''This option is only used when starting a new analysis from .FNA or .FASTA files

 This allows you to set the new names for sequences that MetaCom will save with.
 Each new sequence name will be <IDTag>### with the specified ID tag from this
 option, followed by an incrementing number to distinquish sequences.
 
 Sequences are not numbered in any particular order, other than their apparence
 in the .FASTA or .FNA file you select. 
 
 NOTE: Any sequences that contain EXACTLY identical sequences (including identical
 lengths of the sequences) will be combined into one record with the new name. This
 will greatly decrease overall processing time, and the number of records merged into
 one new name will be recorded with the data as "Copies" to allow analysis based on 
 the original number of input sequences from the .FASTA or .FNA file.
 
 NOTE 2: All original sequence names will be preserved in the table "OldNewNames" for
 ease of reference.'''
        self.lblIDTag  = wx.StaticText(parent=self, id=-1, label="New Prefix for Combined Sequences:")
        self.tbxIDTag  = wx.TextCtrl(parent=self, id=-1, value=options.strIDTag, size = (112,22))
        self.infoIDTag = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strIDTagTip)
        self.sizerIDTag  = self.AddItemsToHorizontalSizer(None, self.infoIDTag, self.lblIDTag , self.tbxIDTag)      

        strMultipleComputerAnalysisTip = '''As larger datasets may take very long times to process on one computer
due to the limit of 2 simultaneous sequence queries per IP address
that BLAST imposes, this option allows for easy division of the 
workload between multiple computers.

All computers will need to have a seperate copy of the source datafile,
to prevent issues occuring with file reading and writing permissions.

It is highly recommended that each computer have a different name on
their project name, in order to avoid issues of sharing filenames and
accidently overwriting files. 

After analysis with this method, the "Combine SQL Files" function can
recombine all result files into one database, and "Check for
Unprocessed Sequences" can insure that all sequences were analyzed.'''
        self.lblMultipleComputerAnalysis  = wx.StaticText(parent=self, id=-1, label="Multiple Computer Analysis:")        
        self.chkMultipleComputerAnalysis = wx.CheckBox(parent=self, id=-1, size=(20,20), style=wx.CHK_2STATE)
        self.chkMultipleComputerAnalysis.SetValue(options.boolMultipleComputerAnalysis)
        self.lblMultipleComputerAnalysis.SetFont(self.FontSubTitle)
        self.infoMultipleComputerAnalysis = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strMultipleComputerAnalysisTip)
        self.sizerMultipleComputerAnalysis = self.AddItemsToHorizontalSizer(self.lblMultipleComputerAnalysis, self.chkMultipleComputerAnalysis, self.infoMultipleComputerAnalysis) 

        strlMultipleComputerNumbersTip = '''These are the settings for multiple computer analyses. If the
option to use multiple computers is unchecked, these values
are not used for anything. 

The first value is the number of the computer in the set that
identifies this computer. This computer will process the 
sequences for that portion of the dataset only. Valid values
are integers >= 1.

The second value is the total number of computers running this
analysis on this file. Each computer will have 1/<this value>
of the dataset to analyze, with any remaining sequences from
this division being processed by the last computer in the set.

For example if the values were 3 and 6, this computer would
process the sequences from 2/6 to 3/6 of the way through the
remaining ones to be done. (The first computer would run
the sequences from 0/6 to 1/6 through the file, etc).'''
        self.lblMultipleComputerNumbers  = wx.StaticText(parent=self, id=-1, label="This Computer's Number:")
        self.tbxMultipleComputerNumbers  = wx.TextCtrl(parent=self, id=-1, value=str(options.intMultipleComputerIndex), size=(58, IMAGE_HEIGHT_PX))
        self.lblMultipleComputerNumbers2  = wx.StaticText(parent=self, id=-1, label=" out of :")
        self.tbxMultipleComputerNumbers2 = wx.TextCtrl(parent=self, id=-1, value=str(options.intMultipleComputerTotalNumber),
                                                       size=(58, IMAGE_HEIGHT_PX))        
        self.infoMultipleComputerNumbers = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strlMultipleComputerNumbersTip)
        self.sizerMultipleComputerNumbers  = self.AddItemsToHorizontalSizer(None, self.infoMultipleComputerNumbers, self.lblMultipleComputerNumbers,
                                                                            self.tbxMultipleComputerNumbers, self.lblMultipleComputerNumbers2,
                                                                            self.tbxMultipleComputerNumbers2)
        
        strBlastAnalysisOptionsTitleTip = '''These options allow you to customize the BLAST analysis run with
parameters that are important to your analysis. 

Please read all warnings to ensure that your result files don't
become larger than expected, and that your IP address doesn't end
up being blocked by the BLAST website due to over submitting.'''
        self.lblBlastAnalysisOptionsTitle = wx.StaticText(parent=self, id=-1,  label="BLAST Analysis Options:")
        self.lblBlastAnalysisOptionsTitle.Font = self.FontSubTitle
        self.infoBlastAnalysisOptionsTitle = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strBlastAnalysisOptionsTitleTip)
        self.sizerBlastAnalysisOptionsTitle = self.AddItemsToHorizontalSizer(self.lblBlastAnalysisOptionsTitle, self.infoBlastAnalysisOptionsTitle)        

        strMinAlignmentTip = '''This sets the minimum accepted alignment length for BLAST results. If no results
satisfy this condition, the sequence read will be recorded as having "No 
Signifcant Similarity Found", even if results are found with shorter alignment
lengths. This alignment is based on the number of base pairs that BLAST actually
aligns between the query and the match sequence, and not on the length of the
query sequence submitted (usually only a sub-region is matched by BLAST).

Leaving this value at the default of 1 will give the best perfomance, results can
always be checked by hand later for more stringent alignment length cutoffs, as 
the alignment lengths are saved with each sequence read's data.'''
        self.lblMinAlignment = wx.StaticText(parent=self, id=-1,  label="Minimum Alignment Length:")
        self.tbxMinAlignment = wx.TextCtrl(parent=self, id=-1,value=str(options.intBLASTMinimumAlignmentLength), size = (163,22))
        self.infoMinAlignment = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strMinAlignmentTip)        
        self.sizerMinAlignment = self.AddItemsToHorizontalSizer(None, self.infoMinAlignment, self.lblMinAlignment, self.tbxMinAlignment)        
        
        strEvalueCutoffExponentTip = '''This sets the minimum accepted E-value on BLAST results. If no results
satisfy this condition, the sequence read will be recorded as having
"No Significant Similarity Found", even if results are found with less
significant E-values. 

The value entered is the EXPONENT of the 1e^-X term, so entering a 6
would mean the E-value must be less than or equal to 1e^-6 to be
included in the result set.

Leaving this value at the default of 6 will give the best perfomance,
results can always be checked by hand later for more stringent E-value
cutoffs, as the E-values are saved with each sequence read's data.'''
        self.lblEvalueCutoffExponent = wx.StaticText(parent=self, id=-1,  label="E-value Cutoff 1e^-X, where X=:")
        self.tbxEvalueCutoffExponent = wx.TextCtrl(parent=self, id=-1,value=str(options.intBLASTEvalueCutoffExponent), size = (132,22))
        self.infoEvalueCutoffExponent = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strEvalueCutoffExponentTip)        
        self.sizerEvalueCutoffExponent = self.AddItemsToHorizontalSizer(None,self.infoEvalueCutoffExponent, self.lblEvalueCutoffExponent,
                                                                        self.tbxEvalueCutoffExponent)            
        
        strMaxResultsPerSequenceTip = '''This sets the number of results that BLAST will return in your query.
        
The default of 250 will usually be sufficient to capture all of the hits
that tie for the highest BLAST Match Score, but performance doesn't 
decrease significantly with values up to 1000.

Lower values may speed up response times by BLAST, but could result
in some tied hits not being returned by BLAST.'''
        strMaxResultsPerSequenceWarning = '''WARNING: Checking this option will likely increase the file size
of your results file as more data will be saved.

This can greatly increase the file size if coupled with saving
all results and uncultured results.

If these other options aren't selected, higher values of this
setting will result in slower response times by the BLAST 
database, but shouldn't appreciably increase the filesize.

This option may also increase the time required to process each
sequence read, as more results must be written to disk.'''
        self.lblMaxResultsPerSequence = wx.StaticText(parent=self, id=-1, label="Max Number of BLAST Hits Saved per sequence:")
        self.cbxMaxResultsPerSequence= wx.ComboBox(parent=self, id=-1, value=cbxlist_MaxResults[options.indexBLASTMaxResultsPerSequence],
                                                   style=wx.CB_READONLY,choices=cbxlist_MaxResults)
        self.infoMaxResultsPerSequence = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strMaxResultsPerSequenceTip)
        self.warningMaxResultsPerSequence = NotifyPanel(parent=self, id=-1, notifyType="warning", toolTipMessage=strMaxResultsPerSequenceWarning)
        self.sizerMaxResultsPerSequence = self.AddItemsToHorizontalSizer(self.warningMaxResultsPerSequence, self.infoMaxResultsPerSequence,
                                                                         self.lblMaxResultsPerSequence, self.cbxMaxResultsPerSequence)
        
        strSimultaneousThreadsTip = '''This option sets how many threads are created to submit sequences to the
BLAST servers at one time (one sequence per thread at a time). Set this
to 1 if you computer isn't powerful enough to run multiple threads. 

Please read the warning on this option, as failure to heed it can result
in blocked connection by the BLAST server.'''
        strSimultaneousThreadsWarning = '''CRITICAL WARNING: BLAST scripting rules dictate that each connected IP
may only submit 2 sequence queries simultaneously. 

Increasing this beyond this runs the risk of having your IP blocked
by the BLAST site as a potential risk and for high-usage. 

For smaller datasets, it may increase the speed of analysis, but for
larger datasets it is HIGHLY ADVISED to use multiple computers rather
than increasing this value.'''
        self.lblSimultaneousThreads = wx.StaticText(parent=self, id=-1,  label="Number of Simultaneous BLAST Thread to run:")
        self.tbxSimultaneousThreads = wx.TextCtrl(parent=self, id=-1,value=str(options.intBLASTSimultaneousThreads), size = (70,22))
        self.infoSimultaneousThreads = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strSimultaneousThreadsTip)
        self.warningSimultaneousThreads = NotifyPanel(parent=self, id=-1, notifyType="warning", toolTipMessage=strSimultaneousThreadsWarning)        
        self.sizerSimultaneousThreads = self.AddItemsToHorizontalSizer(self.warningSimultaneousThreads, self.infoSimultaneousThreads,
                                                                       self.lblSimultaneousThreads, self.tbxSimultaneousThreads )
        
        strSaveAllHitsTip = '''Checking this option will create an additional table in the database
that will store all of the results from the BLAST query, in addition to
the default program behavior, where only the hits that tie for the 
highest BLAST Match Score are saved.

If these results are not of interest to you, do not check this as it
can greatly increase file size.'''
        strSaveAllHitsWarning = '''WARNING: Checking this option will GREATLY increase the file size
of your results file (by up to 10,000x), particularly if you
increased the number of maximum results to include per sequence. 

On larger datasets, this can result in total file sizes easily
in excess of 20GB.

This option may also increase the time required to process each
sequence read.'''
        self.lblSaveAllHits = wx.StaticText(parent=self, id=-1,  label="Include ALL BLAST result hits in save file:")
        self.chkSaveAllHits = wx.CheckBox(parent=self, id=-1, size = (20,20), style=wx.CHK_2STATE)
        self.chkSaveAllHits.SetValue(options.boolBLASTSaveAllHits)
        self.infoSaveAllHits = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strSaveAllHitsTip)
        self.warningSaveAllHits = NotifyPanel(parent=self, id=-1, notifyType="warning", toolTipMessage=strSaveAllHitsWarning)          
        self.sizerSaveAllHits  = self.AddItemsToHorizontalSizer(self.warningSaveAllHits, self.infoSaveAllHits, self.lblSaveAllHits, self.chkSaveAllHits) 
        
        strSaveUnculturedClonesTip = '''Checking this option will create an additional table in the database
that will store all of the results that are from "uncultured" samples.

If these samples are not of interest to you, do not check this as it
can greatly increase file size, and very little useful data will be 
gained from including these sequences.'''
        strSaveUnculturedClonesWarning = '''WARNING: Checking this option will GREATLY increase the file size
of your results file (by up to 10,000x), particularly if you
increased the number of maximum results to include per sequence.

On larger datasets, this can result in total file sizes easily
in excess of 20GB.

This option may also increase the time required to process each
sequence read.'''
        self.lblSaveUnculturedClones = wx.StaticText(parent=self, id=-1,  label="Include ALL Uncultured BLAST result hits in save file:")
        self.chkSaveUnculturedClones = wx.CheckBox(parent=self, id=-1, size = (20,20), style=wx.CHK_2STATE)
        self.chkSaveUnculturedClones.SetValue(options.boolBlastSaveUnculturedClones)
        self.infoSaveUnculturedClones = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strSaveUnculturedClonesTip)
        self.warningSaveUnculturedClones = NotifyPanel(parent=self, id=-1, notifyType="warning", toolTipMessage=strSaveUnculturedClonesWarning)
        self.sizerSaveUnculturedClones  = self.AddItemsToHorizontalSizer(self.warningSaveUnculturedClones, self.infoSaveUnculturedClones,
                                                                         self.lblSaveUnculturedClones, self.chkSaveUnculturedClones) 
         
        self.btnBLASTStart = wx.Button(parent=self, id=-1, label="Begin/Resume BLAST Analysis", size=(200,22))
        self.btnBLASTStart.Bind(wx.EVT_BUTTON, self.OnButtonStart_ClickHandler)
        self.sizerBLASTStart = self.AddItemsToHorizontalSizer(None, None, None, self.btnBLASTStart)
        
        self.sizerLeft = self.AddItemsToVerticalSizer(self.sizerBlastAnalysisTitle,
                                                       self.sizerProject,
                                                       self.sizerEmail,
                                                       self.sizerIDTag,
                                                       self.sizerMultipleComputerAnalysis,
                                                       self.sizerMultipleComputerNumbers,
                                                       self.sizerBlastAnalysisOptionsTitle,
                                                       self.sizerMinAlignment,
                                                       self.sizerEvalueCutoffExponent,
                                                       self.sizerMaxResultsPerSequence,
                                                       self.sizerSimultaneousThreads,
                                                       self.sizerSaveAllHits,
                                                       self.sizerSaveUnculturedClones,
                                                       self.sizerBLASTStart)

        strPostprocessingTitleTip = '''These functions provide you with a range of additional options and results
to help you analyze your data. Some functions require you to have used 
another function on this list before you can use them, as several depend
upon the results of another function to perform their analysis,'''
        self.lblPostprocessingTitle = wx.StaticText(parent=self, id=-1,  label="Additional Analysis and Post-Processing Options:")
        self.lblPostprocessingTitle.Font = self.FontTitle
        self.infoPostprocessingTitle = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strPostprocessingTitleTip)
        self.sizerPostprocessingTitle = self.AddItemsToHorizontalSizer(self.lblPostprocessingTitle, self.infoPostprocessingTitle)
        
        strSequenceTagSummaryTip = '''This generates a table that contains a list of the sequences that MetaCom
identified as being the sequence tag for the sample (located in the first
10 bp of the sequence read). This table can be used to determine what 
groupings your results should have when compared to your primer sequences.

NOTE: some next generation sequencing platforms will read bi-directionally,
so the tag may match to your reverse primer as well.'''
        self.btnSequenceTagSummary = wx.Button(parent=self, id=-1, label="Get Sequence Tag Summary", size = (200,22))
        self.btnSequenceTagSummary.Bind(wx.EVT_BUTTON, self.OnButtonGetSequenceTagSummary_ClickHandler)
        self.infoSequenceTagSummary = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strSequenceTagSummaryTip)        
        self.sizerSequenceTagSummary = self.AddItemsToHorizontalSizer(None, self.infoSequenceTagSummary, self.btnSequenceTagSummary)
        
        strCombineSQLTip = '''This function will take the input SQL files and combine all records in each
table they possess. This result will be saved to a NEW file (any existing 
file with the name you provide for the new file will be erased).

WARNING: This could take a VERY long time if the files are very large, which
is the case if you selected to include large number of sequence results or
uncultured clone hits in your result dataset.'''
        self.btnCombineSQL = wx.Button(parent=self, id=-1, label="Combine SQL Files", size = (200,22))
        self.btnCombineSQL.Bind(wx.EVT_BUTTON, self.OnButtonCombineSQL_ClickHandler)
        self.infoCombineSQL = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strCombineSQLTip)
        self.sizerCombineSQL = self.AddItemsToHorizontalSizer(None, self.infoCombineSQL, self.btnCombineSQL)
        
        strCheckSkippedSequencesTip = '''This function will check your BLAST result table and compare the 
sequences that were analyzed with the entire set of sequences to
run. It will report if any sequences have been skipped or not
processed, and will generate the SequenceData table listing their
information. 

NOTE: this function is for your reference, as clicking "Start/Resume
BLAST Analysis" will automatically call this same function to get the 
sequences to run, so if sequences are skipped, just rerun the BLAST.

NOTE 2: Sequences may be skipped for a variety of reasons, usually
due to internet connection instability at the time of analysis or
improperly set Multiple Computer Analyses. Some sequences, however,
will return errors from BLAST whenever they are used, and so if a
sequence remains on this result list, you may want to check it by
hand.'''
        self.btnCheckSkippedSequences = wx.Button(parent=self, id=-1, label="Check For Unprocessed Sequences", size = (200,22))
        self.btnCheckSkippedSequences.Bind(wx.EVT_BUTTON, self.OnButtonCheckSkippedSequences_ClickHandler)
        self.infoCheckSkippedSequences = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strCheckSkippedSequencesTip)
        self.sizerCheckSkippedSequences = self.AddItemsToHorizontalSizer(None, self.infoCheckSkippedSequences, self.btnCheckSkippedSequences)        
        
        strGetTaxonomyTip = '''This analysis will get the taxonomic assignment for the match based on the ASC
number from the blast result. The ASC number is used to look up the species ID
in GenBank, and from there the taxonomic classification is read from NCBI's 
taxonomy database.
                               
Due to the variety of different gradations of levels present in the NCBI
assigment (subfamily, superfamily, etc), only the levels (as defined by NCBI)
of Superkingdom/Domain, Phylum, Class, Order, Family, Genus, Species are 
collected.'''
        self.btnGetTaxonomy = wx.Button(parent=self, id=-1, label="Get Taxonomy Assignments", size = (200,22))
        self.btnGetTaxonomy.Bind(wx.EVT_BUTTON, self.OnButtonGetTaxonomy_ClickHandler)
        self.infoGetTaxonomy = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strGetTaxonomyTip)
        self.sizerGetTaxonomy = self.AddItemsToHorizontalSizer(None, self.infoGetTaxonomy, self.btnGetTaxonomy)
        
        strGetRecommendedHitsTip = '''This analysis should be conducted only AFTER getting all taxonomy assignments,
by using the "Get Taxonomy Assignments" button.
                                      
This will compare the taxonomic assignments for each match that shares the
top BLAST result score (in case of ties), and then will assign each taxonomic
level the consensus assignment for that level. Levels that have no assignment
indicate that the species assignments disagreed at this level, and thus no
consensus assignment could be made with confidence at this level.
                                      
For example if you had two hits tying for the top BLAST alignment score of:
"Eukaryota, Chordata, Mammalia, Fibrobacterales, Bovidae, Bos, taurus" and
"Eukaryota, Chordata, Mammalia, Fibrobacterales, Bovidae, Capra, hircus".
The resulting classification would be (where <none> is actually no text):
Eukaryota, Chordata, Mammalia, Fibrobacterales, Bovidae, <none>, <none>
since the two records only have a consensus assignment to the Bovidae family.
                                      
The RecommendedHit table will also include the name of the sequence record,
the sequence's multiplex identifier tag (MID) or the first 10 bp if no MID,
the count of how many identical original reads were combined into that record,
the BLAST score of the records match, and the full sequence of the record.'''
        self.btnGetRecommendedHits = wx.Button(parent=self, id=-1, label="Get Recommended Species Hits", size = (200,22))
        self.btnGetRecommendedHits.Bind(wx.EVT_BUTTON, self.OnButtonGetRecommendedHits_ClickHandler)
        self.infoGetRecommendedHits = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strGetRecommendedHitsTip)
        self.sizerGetRecommendedHits = self.AddItemsToHorizontalSizer(None, self.infoGetRecommendedHits, self.btnGetRecommendedHits)
        
        strGetPercentIdentityTip = '''This generates a table summarizing the percent identity of the all of the top
scoring BLAST results (that were tied for highest score) used to assign the
taxonomic levels to the sequence.

The output columns are the sequence name of the sequence record, the name of
hit returned by BLAST, the percent identity (expressed from 0.00 to 1.00),
and the length of the sequence that was aligned to the BLAST result.'''
        self.btnGetPercentIdentity = wx.Button(parent=self, id=-1, label="Get Percent Identity Summary", size = (200,22))
        self.btnGetPercentIdentity.Bind(wx.EVT_BUTTON, self.OnButtonGetPercentIdentity_ClickHandler)
        self.infoGetPercentIdentity = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strGetPercentIdentityTip)
        self.sizerGetPercentIdentity = self.AddItemsToHorizontalSizer(None, self.infoGetPercentIdentity, self.btnGetPercentIdentity)
        
        strTaxonomySummaryTitleTip = '''These functions will generate taxonomy summaries for your data, either at a
single level or including the entire classification down to that level in the
full report. 

Due to variablity in how many gradations of a particular level each species
is annotated with (subspecies, superspecies, etc) only the traditional levels
of taxonomy are listed (Kingdom/Superkingdom, Phylum, Class, Order, Family,
Genus, Species).
                                        
Records with blank names indicate the number of records that were not able to
be classified at that level, due to several BLAST hits that had different 
classifications at that level tying in score.'''
        self.lblTaxonomySummaryTitle = wx.StaticText(parent=self, id=-1,  label="Taxonomy Summary Options:")
        self.lblTaxonomySummaryTitle.Font = self.FontSubTitle
        self.infoTaxonomySummaryTitle = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strTaxonomySummaryTitleTip)
        self.sizerTaxonomySummaryTitle = self.AddItemsToHorizontalSizer(self.lblTaxonomySummaryTitle, self.infoTaxonomySummaryTitle)      
        
        strFullTaxonomyReportTip = '''This will generate the taxonomy summary for your data, with all levels from
superkingdom down to the level selected listed. Due to variablity in how many
gradations of a particular level each species is annotated with (subspecies, 
superspecies, etc) only the traditional levels of taxonomy are listed
(Kingdom/Superkingdom, Phylum, Class, Order, Family, Genus, Species).
                                      
For example, if you select species and had human DNA in your sample, it will
give each of the major levels for humans, followed by a count of the total 
number of sequences that were assigned down to the species level.
                                      
The record with the blank name indicates the number of records that were not
able to be classified at this level, due to several BLAST hits that had 
different classifications at this level tying in score.'''
        self.lblFullTaxonomyReport = wx.StaticText(parent=self, id=-1, label="Output Full Taxonomy Report Down to Level:")
        self.cbxFullTaxonomyReport = wx.ComboBox(parent=self, id=-1, value=cbxlist_TaxonomyLevels[options.indexTaxonomyFullOutputLevel],style=wx.CB_READONLY,choices=cbxlist_TaxonomyLevels)
        self.btnFullTaxonomyReport = wx.Button(parent=self, id=-1, label="Generate Report", size = (114, 22))
        self.btnFullTaxonomyReport.Bind(wx.EVT_BUTTON, self.OnButtonFullTaxonomyReport_ClickHandler)        
        self.infoFullTaxonomyReport = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strFullTaxonomyReportTip)
        self.sizerFullTaxonomyReport = self.AddItemsToHorizontalSizer(None, self.infoFullTaxonomyReport, self.lblFullTaxonomyReport,
                                                                      self.cbxFullTaxonomyReport, self.btnFullTaxonomyReport)
        
        strSingleTaxonomyReportTip = '''This will generate the taxonomy summary for your data for a single level. 

For example, if you select kingdom (which using NCBI standards is actually superkingdom),
it will give the count of Bacteria, Archea, Eurkaryota in the sample.

The record with the blank name indicates the number of records that were not able to be
classified at this level, due to several BLAST hits that had different classifications
at this level tying in score.'''
        self.lblSingleTaxonomyReport = wx.StaticText(parent=self, id=-1, label="Output Single Level Taxonomy Report for Level:")
        self.cbxSingleTaxonomyReport= wx.ComboBox(parent=self, id=-1, value=cbxlist_TaxonomyLevels[options.indexTaxonomySingleOutputLevel],style=wx.CB_READONLY,choices=cbxlist_TaxonomyLevels)
        self.btnSingleTaxonomyReport = wx.Button(parent=self, id=-1, label="Generate Report")
        self.btnSingleTaxonomyReport.Bind(wx.EVT_BUTTON, self.OnButtonSingleTaxonomyReport_ClickHandler)        
        self.infoSingleTaxonomyReport = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strSingleTaxonomyReportTip)
        self.sizerSingleTaxonomyReport = self.AddItemsToHorizontalSizer(None, self.infoSingleTaxonomyReport, self.lblSingleTaxonomyReport,
                                                                        self.cbxSingleTaxonomyReport, self.btnSingleTaxonomyReport)         
        
        strGetRarefactionCurvesTip = '''This will generate the data values needed to plot the rarefaction curve for your
dataset on a graph, and save it to the SQL database. 100 randomly samplings of 
the dataset are used to generate these datapoints. 

The SampleNumber column will be your X-axis values.

The remaining columns your Y-axis values. They represent the average number of
unique taxonomic names drawn from the pool of all sequence reads at that sample
number.

All taxonomy values are taken from the RecommendedHit table, and each record in
that table is entered into the pool once for each copy that it represents.

For example, if there were three perfectly identical sequences in the original
file, they represent only one record in RecommendedHit, but there will be three
entries with their taxonomic assignment in the sample pool for this function.'''
        self.btnGetRarefactionCurves = wx.Button(parent=self, id=-1, label="Get Rarefaction Curves", size = (200,22))
        self.btnGetRarefactionCurves.Bind(wx.EVT_BUTTON, self.OnButtonGetRarefactionCurves_ClickHandler)
        self.infoGetRarefactionCurves = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strGetRarefactionCurvesTip)       
        self.sizerGetRarefactionCurves = self.AddItemsToHorizontalSizer(None, self.infoGetRarefactionCurves, self.btnGetRarefactionCurves)        

        strGetShannonIndexTip = '''This will calculate the Shannon Index for your dataset and will output the results both
to a table in the SQL database and will display the result onscreen. The table will
include all intermediate values for the proportion, ln(proportion), and the product of
the two, to allow for checking by hand, if desired.

Values for the Shannon Index will be calculated using two sets of species counts, one
including the records that have either genus or species unassigned, and one including
only records that had an identified genus and species.'''
        self.btnGetShannonIndex = wx.Button(parent=self, id=-1, label="Get Shannon Index", size = (200,22))
        self.btnGetShannonIndex.Bind(wx.EVT_BUTTON, self.OnButtonGetShannonIndex_ClickHandler)
        self.infoGetShannonIndex = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strGetShannonIndexTip)   
        self.sizerGetShannonIndex = self.AddItemsToHorizontalSizer(None, self.infoGetShannonIndex, self.btnGetShannonIndex) 
        
        strExportToExcelTip = '''This button will take all current SQL tables and views and save them each to a seperate Excel
file in the same folder as the source SQL file. If you don't see the table you are looking for,
make sure that you have generated it with the correct analysis button first.'''
        self.btnExportToExcel = wx.Button(parent=self, id=-1, label="Export all SQL tables to Excel Files", size = (200,22))
        self.btnExportToExcel.Bind(wx.EVT_BUTTON, self.OnButtonExportToExcel_ClickHandler)
        self.infoExportToExcel = NotifyPanel(parent=self, id=-1, notifyType="info", toolTipMessage=strExportToExcelTip)   
        self.sizerExportToExcel = self.AddItemsToHorizontalSizer(None, self.infoExportToExcel, self.btnExportToExcel) 
        
        self.sizerRight = self.AddItemsToVerticalSizer(self.sizerPostprocessingTitle,
                                                       self.sizerSequenceTagSummary,
                                                       self.sizerCombineSQL,
                                                       self.sizerCheckSkippedSequences, 
                                                       self.sizerGetTaxonomy,
                                                       self.sizerGetRecommendedHits,
                                                       self.sizerGetPercentIdentity,
                                                       self.sizerGetRarefactionCurves,
                                                       self.sizerGetShannonIndex,                                                         
                                                       self.sizerExportToExcel,
                                                       self.sizerTaxonomySummaryTitle,                                          
                                                       self.sizerFullTaxonomyReport,
                                                       self.sizerSingleTaxonomyReport)

        self.sizerCombined = self.AddItemsToHorizontalSizer(self.sizerLeft, self.sizerRight)       
        self.SetSizer(self.sizerCombined)
        self.Centre()    
        
        # Set up event handler for any worker thread results
        EVT_UPDATESTATUS(self, self.OnUpdateStatus)
        pub.subscribe(self._OnObjectAdded, 'object.added')

        self.CreateStatusBar()
        wx.PostEvent(self, ResultEvent("Select Analysis Options", EVT_UPDATESTATUS_ID))
        
        options.SaveOptions()

        self.SetAutoLayout(True)
        self.Layout()
        self.Show()
    
    def AddItemsToHorizontalSizer(self, *args):
        defaultBorderSize = 8
        defaultHorizontalSpacing = 8
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        sizer.AddSpacer(defaultBorderSize)
        for itemtoadd in args:
            if itemtoadd is None:
                sizer.AddSpacer(IMAGE_WIDTH_PX)
            else: 
                weight = 0
                flag = 0
                if (type(itemtoadd) is wx.TextCtrl) or (type(itemtoadd) is wx.ComboBox) or (type(itemtoadd) is wx.Sizer):
                    weight = 1
                    flag = wx.ALIGN_CENTER_VERTICAL
                sizer.Add(itemtoadd, weight, flag)
            sizer.AddSpacer(defaultHorizontalSpacing)
        sizer.AddSpacer(defaultBorderSize)
            
        return sizer
        
    def AddItemsToVerticalSizer(self, *args):
        defaultBorderSize = 10
        defaultVerticalSpacing = 10
        sizer = wx.BoxSizer(wx.VERTICAL)
        
        sizer.AddSpacer(defaultBorderSize)
        for itemtoadd in args:
            sizer.Add(itemtoadd, 0, 0, 0, None)
            sizer.AddSpacer(defaultVerticalSpacing)
        sizer.AddSpacer(defaultBorderSize)

        return sizer   
     
    def OnUpdateStatus(self, event):
        '''Updates the status bar with the supplied message.'''
        self.StatusBar.SetStatusText(event.data)
        self.Refresh()
        
    def _OnObjectAdded(self, message=""):
        self.StatusBar.SetStatusText(message)

        
    def GetOptions(self):
        '''Gets the options from the fields on the supplied frame. Also opens
        a file selection dialog box that allows multiple .fna files to be
        selected. Returns the options object and a boolean variable indicating
        if a file was selected.'''
        bool_file_selected = False
    
        #Open File Browser
        self.dlg = wx.FileDialog(self,
                            message="Select File to analyze",
                            wildcard="454 and FASTA Files (.fna, .fa, .fasta, .sqlite, .txt)|*.fna;*.fa;*.fasta;*.sqlite;*.txt|All files (*.*)|*.*",
                            style=wx.FD_OPEN)
        self.dlg.ShowModal()
    
        try:
            if self.dlg.Paths:
                bool_file_selected = True
                paths = self.dlg.Paths
                FolderForOutput = os.path.dirname(paths[0])
                options = self.ReadOptionsFromForm(paths)
                
                #Check for some issues in options
                if options.intMultipleComputerIndex < 1:
                    options.intMultipleComputerIndex = 1
                if options.intMultipleComputerTotalNumber < 1:
                    options.intMultipleComputerTotalNumber = 1
                if options.intMultipleComputerIndex > options.intMultipleComputerTotalNumber:
                    options.intMultipleComputerIndex = options.intMultipleComputerTotalNumber
                         
                options.SaveOptions()
            else:
                bool_file_selected = False
                options = Options()            
        except:
            bool_file_selected = False
            options = Options()
        
        self.dlg.Destroy()
        
        return options, bool_file_selected
    
    def ReadOptionsFromForm(self, filepaths):
        '''Reads the options from the form, and returns the options object'''
        options = Options(fna_filepaths=filepaths,
                          strProjectName=self.tbxProject.GetValue(),
                          strIDTag=self.tbxIDTag.GetValue(),
                          boolMultipleComputerAnalysis=self.chkMultipleComputerAnalysis.IsChecked(),
                          intMultipleComputerIndex = int(self.tbxMultipleComputerNumbers.GetValue()),
                          intMultipleComputerTotalNumber=int(self.tbxMultipleComputerNumbers2.GetValue()),
                          intBLASTMinimumAlignmentLength=int(self.tbxMinAlignment.GetValue()),
                          intBLASTEvalueCutoffExponent=int(self.tbxEvalueCutoffExponent.GetValue()),                                                           
                          indexBLASTMaxResultsPerSequence=self.cbxMaxResultsPerSequence.GetCurrentSelection(),
                          intBLASTSimultaneousThreads=int(self.tbxSimultaneousThreads.GetValue()),
                          boolBLASTSaveAllHits=self.chkSaveAllHits.IsChecked(),
                          boolBlastSaveUnculturedClones=self.chkSaveUnculturedClones.IsChecked(),
                          indexTaxonomyFullOutputLevel=self.cbxFullTaxonomyReport.GetCurrentSelection(),
                          indexTaxonomySingleOutputLevel=self.cbxSingleTaxonomyReport.GetCurrentSelection())
        return options
    
    def OnButtonStart_ClickHandler(self, event):
        '''Starts analysis on button click'''
        self.StartThread(parent=self, taskToRun="CollectData", event=event)
        
    def OnButtonCombineSQL_ClickHandler(self, event):
        '''Starts prompt for combining files on button click'''
        self.StartThread(parent=self, taskToRun="CombineSQLFiles", event=event)       
        
    def OnButtonCheckSkippedSequences_ClickHandler(self, event):
        '''Calls the function to remove duplicate entries'''
        self.StartThread(parent=self, taskToRun="FindSkippedSequences", event=event) 
        
    def OnButtonGetRecommendedHits_ClickHandler(self, event):
        '''Goes to OnTaxonomyResult Function'''     
        self.StartThread(parent=self, taskToRun="GetRecommendedHits", event=event)
        
    def OnButtonGetSequenceTagSummary_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="GetSequenceTagSummary", event=event)

    def OnButtonGetPercentIdentity_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="GetPercentIdentity", event=event)
        
    def OnButtonFullTaxonomyReport_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="FullTaxonomyReport", event=event)
   
    def OnButtonSingleTaxonomyReport_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="SingleTaxonomyReport", event=event)
        
    def OnButtonGetRarefactionCurves_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="GetRarefactionCurves", event=event)
        
    def OnButtonGetShannonIndex_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="GetShannonIndex", event=event)
  
    def OnButtonExportToExcel_ClickHandler(self, event):
        self.StartThread(parent=self, taskToRun="ExportToExcel", event=event)
        
    def OnButtonGetTaxonomy_ClickHandler(self, event):
        '''Starts analysis on button click'''
        self.StartThread(parent=self, taskToRun="GetTaxonomy", event=event)

    def StartThread(self, parent, taskToRun, event):           
        SideThread = ProcessingThread(parent=self, taskToRun=taskToRun, event=event)
        SideThread.start()
                        
        
class ProcessingThread(threading.Thread):
    def __init__(self, parent, taskToRun, event):
        '''A worker thread that takes the information from the main window, and executes the code
        seperately, to prevent the main window from hanging. This is the general thread class that
        will get called for any button or processing sent from the main page.'''
        
        threading.Thread.__init__(self)
        
        self.setDaemon(True)
        self.parent = parent
        self.options = None
        self.bool_file_selected = False
        self.taskToRun = taskToRun
        sql_filepath = ""                
        
        if not self.taskToRun == "CombineSQLFiles":
            self.options, self.bool_file_selected = parent.GetOptions()
        else:
            self.options = Options([])
            self.bool_file_selected = False
            
            self.options.SaveOptions()
            
        self.event = event
        
        self.boolFileTypeError = False
        
        if type(self.options.fna_filepaths) == list:
            if not self.options.fna_filepaths == []:
                self.sql_filepath = self.options.fna_filepaths[0]
            else:
                self.sql_filepath = ""
        elif type(self.options.fna_filepaths) == str:
            self.sql_filepath = self.options.fna_filepaths
        else:
            self.boolFileTypeError = True
            wx.MessageBox("Error in DNA source file selection", "Error")
            
        if self.sql_filepath.endswith(".sqlite"):
            self.bool_sql_file_valid = True
        else:
            self.bool_sql_file_valid = False      
         
    def run(self):
        #Throw the event that was passed to it again, and handle it here
        if not self.boolFileTypeError:
            self.RunCorrectTask()
    
    def RunCorrectTask(self):
        '''This function determines what task to run, and calls its correct function'''
        if (self.taskToRun == "CollectData"):
            self.CollectData()
        elif (self.taskToRun == "CombineSQLFiles"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True 
                
            self.CombineSQLFiles(self.parent)
        elif (self.taskToRun == "FindSkippedSequences"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True  
                
            count_skipped = FindSkippedSequences(options=self.options, window=parent)
            if count_skipped > 0:
                wx.MessageBox(message = "Sequences in table SequenceData_All without BLAST results saved to table SequenceData", caption="Task Complete", style =wx.OK)
            elif count_skipped == 0:
                wx.MessageBox(message = "No Skipped Sequences Found!", caption="Task Complete", style =wx.OK)
        elif (self.taskToRun == "GetRecommendedHits"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate recommended hits, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                sql_filepath = self.options.fna_filepaths[0]
                self.GetRecommendedHits()
        elif (self.taskToRun == "GetSequenceTagSummary"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate sequence tag summary, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetSequenceTagSummary()         
        elif (self.taskToRun == "GetPercentIdentity"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate percent identities, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetPercentIdentities()
        elif (self.taskToRun == "FullTaxonomyReport"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate Full Taxonomy Report identities, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetFullTaxonomyReport()
        elif (self.taskToRun == "SingleTaxonomyReport"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate Single Taxonomy Report, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetSingleTaxonomyReport()
        elif (self.taskToRun == "GetRarefactionCurves"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate Rarefaction Curves values, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetRarefactionCurves()
        elif (self.taskToRun == "GetShannonIndex"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to generate Shannon Index, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.GetShannonIndex()
        elif (self.taskToRun == "ExportToExcel"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to export SQL tables to Excel files, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                self.ExportAllFiles()
        elif (self.taskToRun == "GetTaxonomy"):
            if not self.bool_sql_file_valid:
                bool_file_selected = self.GetSQLFilePath()
            else: 
                sql_filepath = self.options.fna_filepaths[0]                
                bool_file_selected = True
        
            if (not bool_file_selected):
                wx.MessageBox(message = "Unable to acquire Taxonomy data, no SQLite file selected", caption="No SQLite File Selected.", style =wx.ICON_ERROR)            
            else:
                #Gets the SQL filepath
                self.GetTaxonomy()            

    def GetTaxonomy(self):
        list_raw_ASC = []
        list_ASC = []

        con = dbapi.connect(self.sql_filepath)
        cur = con.cursor()
        
        #Set database output into python to be in regular string format
        con.text_factory = str 
        
        #Create Output Table
        cur.execute('''CREATE TABLE IF NOT EXISTS Taxonomy
            (TaxonomyID INTEGER NOT NULL, ASC VARCHAR, Kingdom VARCHAR,
            Phylum VARCHAR, Class VARCHAR, TaxonOrder VARCHAR, Family VARCHAR,
            Genus VARCHAR, Species VARCHAR,
            PRIMARY KEY (TaxonomyID))''')
        con.commit()
        
        #Get all ASC data from the file
        cur.execute('''SELECT DISTINCT ASC FROM AllTopHits
                       ORDER BY ASC''')
        myquery_ascdata = cur.fetchall()                
        
        for datarow in myquery_ascdata:
            #Add all ASC to list_Seq454
            if not datarow[0] == "":
                list_raw_ASC.append(datarow[0])
        
        #Set Start Position
        current_index = 0
        if self.options.intStartingPosition >= len(list_raw_ASC):
            #Analyze just last sequence
            current_index = len(list_raw_ASC) - 1
        elif self.options.intStartingPosition < 0:
            current_index = 0
        else:
            current_index = self.options.intStartingPosition
            
        #Set list to analyze as sublist of the complete list    
        list_ASC = list_raw_ASC[current_index:]
        
        #Assign email
        email = self.options.strEmail.strip()
        
        list_skipped_ASC = []
        
        for ASC in list_ASC:
            try:
                #Wait 2 seconds to minimize load on server (it allows up to 3 per second, but this is less stressing)
                time.sleep(2)
                
                #Get Taxonomy
                taxonomy = FindTaxonomy(ASC, email)
                                      
                #Save output of taxonomy query
                cur.execute('''INSERT INTO Taxonomy VALUES(?,?,?,?,?,?,?,?,?)''',
                            (None, ASC, taxonomy.get("superkingdom",""),
                             taxonomy.get("phylum",""), taxonomy.get("class",""),
                             taxonomy.get("order",""), taxonomy.get("family",""),
                             taxonomy.get("genus",""), taxonomy.get("species","")))
                con.commit()
            except:
                list_skipped_ASC.append(ASC)
                next
        
        #Retry Missed Sequences once
        for ASC in list_skipped_ASC:
            try:
                #Wait 1 seconds to minimize load on server (it allows up to 3 per second, but this is less stressing)
                time.sleep(1)
                
                #Get Taxonomy
                taxonomy = FindTaxonomy(ASC, email)
                                      
                #Save output of taxonomy query
                cur.execute('''INSERT INTO Taxonomy VALUES(?,?,?,?,?,?,?,?,?)''',
                            (None, ASC, taxonomy.get("superkingdom",""),
                             taxonomy.get("phylum",""), taxonomy.get("class",""),
                             taxonomy.get("order",""), taxonomy.get("family",""),
                             taxonomy.get("genus",""), taxonomy.get("species","")))
                con.commit()
            except:
                next                
            
        wx.MessageBox("Taxonomic Data Collected and stored to table Taxonomy", "Task Complete", style = wx.OK)
        
    def CollectData(self):
        '''Gets the options data from the selected file'''
        self.UpdateStatus("Reading Options")
    
        if self.options.fna_filepaths:
            #Get filepath object, and start analysis in the spot appropriate to that filetype
            file_type = self.options.fna_filepaths[0].split(".")[-1].lower()
            
            list_454 = []
            list_combined_454 = []
            
            if file_type == "fna":
                self.UpdateStatus("Creating FASTA Files")
                fasta_filepath = CreateFASTA_FromFNA(self.options.fna_filepaths, self.options.strProjectName, self)
                
                self.UpdateStatus("Creating Sequence Objects")
                list_454 = ParseFASTA_IntoSeq454(fasta_filepath, self)
            
                self.UpdateStatus("Editing Sequence Tags")
                EditSeq454Tags(list_454, self.options.strProjectName, self)
    
                self.UpdateStatus("Combining Identical Sequences")
                list_combined_454 = CombineAndRenameIdenticalSequences(list_454, self.options.strIDTag, self)
            
                self.UpdateStatus("Saving Sequences Prior To BLAST")
                sql_filepath = SaveCombined454Files(list_combined_454, self.options.strProjectName, self, self.options.fna_filepaths)
                
            elif file_type == "fa" or file_type == "fasta" or file_type == "txt":
                self.UpdateStatus("Creating FASTA Files")
                fasta_filepath = CreateFASTA_FromFASTA(self.options.fna_filepaths, self.options.strProjectName, self)
                    
                self.UpdateStatus("Creating Sequence Objects")
                list_454 = ParseFASTA_IntoSeq454(fasta_filepath, self)
            
                self.UpdateStatus("Editing Sequence Tags")
                EditSeq454Tags(list_454, self.options.strProjectName, self)
    
                self.UpdateStatus("Combining Identical Sequences")
                list_combined_454 = CombineAndRenameIdenticalSequences(list_454, self.options.strIDTag, self)

                self.UpdateStatus("Saving Sequences Prior To BLAST")
                sql_filepath = SaveCombined454Files(list_combined_454, self.options.strProjectName, self, self.options.fna_filepaths) 
                
            elif file_type == "sqlite":
                #self.UpdateStatus("Creating Sequence Objects")
                #list_combined_454 = ParseSQLite_IntoSeq454(self.options.fna_filepaths, self)
                sql_filepath = self.options.fna_filepaths[0]
                
            self.sql_filepath = sql_filepath
            #Force lists free in memory
            list_454 = []
            list_combined_454 = []            

            self.OnDataCollectResult()
         
    def OnDataCollectResult(self):
        if self.sql_filepath.endswith(".sqlite"):
            self.UpdateStatus("Beginning BLAST Analysis...")
            #Data was successfully collected, start BLAST work
            BlastFrame = ConductBLASTAnalysis(self, self.parent, sql_filepath = self.sql_filepath,
                                                   options = self.options)
            BlastFrame.start()
        else:
            wx.MessageBox("Unknown Filetype", "Unknown filetype unable to process", wx.ok)
    
    def GetRecommendedHits(self):
        '''After all taxonomy analysis is done, starts working on the post analysis'''
        con = dbapi.connect(self.sql_filepath)
        cur = con.cursor()
        
        #Set database output into python to be in regular string format
        con.text_factory = str             
        
        #Make the final summary table (it does contain redundant info, but is meant to be output
        #to Excel in completed form
        cur.execute('''CREATE TABLE IF NOT EXISTS RecommendedHit
                    (HitID INTEGER NOT NULL, SeqName VARCHAR NOT NULL, Tag VARCHAR, Copies VARCHAR,
                    Kingdom VARCHAR, Phylum VARCHAR, Class VARCHAR,
                    TaxonOrder VARCHAR, Family VARCHAR, Genus VARCHAR, Species VARCHAR, 
                    MatchScore REAL, CompleteSequence VARCHAR,
                    PRIMARY KEY (HitID))''')
        con.commit()
        
        #Get all ASC data from the file
        cur.execute('''SELECT DISTINCT SeqName FROM AllTopHits
                       WHERE MatchScore != "0"''')
        myquery_seqnames = cur.fetchall()
        
        recommended_counter = 0
        starting_position = 0
        for seqnamerow in myquery_seqnames:
            recommended_counter += 1
            
            #allows for interrupted analyses to restart faster
            if recommended_counter > starting_position:
                #Initialize Variables
                list_TopHits = []
                list_ASC = []
                dict_taxonomy = dict()
                dict_taxonomy = {"":["","","","","","",""]}

                #Gather additional information about the sequence
                if not seqnamerow[0] == "":
                    #Get All datarows from AllTopHits
                    cur.execute('''SELECT DISTINCT ASC, MatchName, MatchScore
                                           FROM AllTopHits WHERE SeqName = "%s"''' % (seqnamerow[0]))
                    myquery_tophits = cur.fetchall()
                    
                    for tophit in myquery_tophits:
                        if not tophit[1] == "":
                            list_TopHits.append(tophit)
                            list_ASC.append(tophit[0])
                    
                    #Get All Sequence Data
                    cur.execute('''SELECT DISTINCT Copies, Tag, CompleteSequence
                                   FROM SequenceData_All WHERE SeqName = "%s"''' % (seqnamerow[0]))
                    myquery_seqdata = cur.fetchall()
                    
                    for seqdata in myquery_seqdata:
                        if not seqdata[2] == "":
                            seqdata_copies = seqdata[0] 
                            seqdata_tag = seqdata[1]
                            seqdata_sequence = seqdata[2]
                        else:
                            wx.MessageBox("Error: No sequence data found for %s" % (seqnamerow[0]))
                    
                    #Get all taxonomy data (try catch escapes whenever ASC value doesn't exist)
                    for ASC in list_ASC:
                        try:
                            if not ASC == "":
                                cur.execute('''SELECT DISTINCT Kingdom, Phylum, Class, TaxonOrder, Family,
                                                               Genus, Species
                                               FROM Taxonomy WHERE ASC = "%s"''' % (ASC))
                                myquery_taxondata = cur.fetchall()
                                dict_taxonomy[ASC] = myquery_taxondata[0]
                        except:
                            next
                    
                    list_FinalTaxonomy = []
                    list_FinalTopHit = []
                    #Traceback the taxonomy data for all ASC numbers to find the consensus
                    if len(list_TopHits) < 1:
                        wx.MessageBox("Error: list_TopHits is empty or invalid for %s" % (seqnamerow[0]))
                    elif len(list_TopHits) == 1:
                        #There is only one top hit, so just pass the values into the table
                        list_FinalTaxonomy = dict_taxonomy.get(list_TopHits[0][0], ["","","","","","",""])
                        list_FinalTopHit = list_TopHits[0]
                    elif len(list_TopHits) > 1:
                        list_taxonomy = []
                        for ASC in list_ASC:
                            list_taxonomy.append(dict_taxonomy.get(ASC,["","","","","","",""]))
                        
                        bool_match = True    
                        for x in range(len(cbxlist_TaxonomyLevels)):
                            for y in range(1, len(list_TopHits)):
                                #Match all levels against the top hit, changing match to false if it isn't a match
                                if not list_taxonomy[y][x] == list_taxonomy[0][x]:
                                    #Goes to false if even one doesn't match
                                    bool_match = False
                                    break
                                    
                            #Write any value to the final taxonomy list if all match, write "" if not match
                            if bool_match == True:
                                list_FinalTaxonomy.append(list_taxonomy[0][x])
                            else:
                                list_FinalTaxonomy.append("")
                        
                        for tophit in list_TopHits:
                            if not tophit[0] == "":
                                list_FinalTopHit = list_TopHits[0] 
                                break           
                        
                    #Get the data from the correct lists
                    (taxondata_kingdom, taxondata_phylum, taxondata_class, taxondata_order, taxondata_family,
                     taxondata_genus, taxondata_species) = list_FinalTaxonomy
                    
                    #Note: ASC and Matchname won't be included in this table
                    (matchdata_ASC, matchdata_matchname, matchdata_matchscore) = list_FinalTopHit
  
                    #Save the consensus data to new table
                    cur.execute('''INSERT INTO RecommendedHit
                                   (HitID, SeqName, Tag, Copies, Kingdom, Phylum, Class, 
                                    TaxonOrder, Family, Genus, Species,
                                    MatchScore, CompleteSequence)
                                   VALUES (NULL,?,?,?,?,?,?,?,?,?,?,?,?)''',
                               (seqnamerow[0], seqdata_tag, seqdata_copies,
                                    taxondata_kingdom, taxondata_phylum, taxondata_class,
                                    taxondata_order, taxondata_family, taxondata_genus,
                                    taxondata_species, matchdata_matchscore, seqdata_sequence))
                    con.commit()                   
        #OutputCompletionMessage
        wx.MessageBox("All Sequence Analysis Complete.\nFiles saved to source file's directory",
                      'Analysis Complete',
                      wx.OK)            
    
    def GetSequenceTagSummary(self):
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "SequenceData_All" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 

            cur.execute('''DROP VIEW IF EXISTS AllTags''')
            con.commit()
            cur.execute('''CREATE VIEW "AllTags" AS SELECT Tag, Sum(Copies) FROM SequenceData_All GROUP BY Tag ORDER BY Sum(Copies) DESC'''
                        % (strLevelToReport, strFullLevels, strFullLevels))
            con.commit()        
            
            wx.MessageBox("Successfully created list of all sequence tags identified", caption = 'Task Complete', style=wx.OK)                          
        else:
            wx.MessageBox("No table named RecommendedHit present in SQL file. Try running 'Get Recommended Hits' first.", caption = 'Error', style=wx.ICON_ERROR) 
    
        
    def GetPercentIdentities(self):
        '''Creates a file with the percent identity of each sequence read'''
       
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "AllTopHits" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 

            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 
            
            cur.execute('''DROP VIEW IF EXISTS Percent_Identities''')
            con.commit()
            cur.execute('''CREATE VIEW Percent_Identities  AS
            SELECT SeqName, MatchName, ((Identities)/(abs(SubjectEnd - SubjectStart) + 1)) AS PercentIdentity, Identities, (abs(SubjectEnd - SubjectStart) +1) AS MatchLength From AllTopHits
            WHERE SubjectStart > 0
            ORDER BY PercentIdentity DESC''')
            con.commit()
            
            excel_filepath = self.OutputExcelFile(self.sql_filepath, "Percent_Identities")                         
        else:
            wx.MessageBox("No table named AllTopHits present in SQL file. Try running 'Begin/Resume BLAST Analysis' first.", caption = 'Error', style=wx.ICON_ERROR) 
            
        
    def GetFullTaxonomyReport(self):
        '''Outputs a taxonomy report that summarizes the taxon level selected and the count for that level'''
        
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "RecommendedHit" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 

            strLevelToReport = cbxlist_TaxonomyLevels[self.options.indexTaxonomyFullOutputLevel]

            list_HigherLevels = cbxlist_TaxonomyLevels[0:self.options.indexTaxonomyFullOutputLevel + 1]
            
            strFullLevels = ""
            for level in list_HigherLevels:
                if level == "Order":
                    level = "TaxonOrder"
                
                if strFullLevels == "":
                    strFullLevels = level
                else:
                    strFullLevels += ", " + level
                

            cur.execute('''DROP VIEW IF EXISTS FullTaxonReport_%s''' % (strLevelToReport))
            con.commit()
            cur.execute('''CREATE VIEW FullTaxonReport_%s AS SELECT %s, Sum(Copies) as Total From RecommendedHit GROUP BY %s ORDER BY Total DESC'''
                        % (strLevelToReport, strFullLevels, strFullLevels))
            con.commit()
            
            cur.execute('''DROP VIEW IF EXISTS FullTaxonReportByTag_%s''' % (strLevelToReport))
            con.commit()
            cur.execute('''CREATE VIEW FullTaxonReportByTag_%s AS SELECT Tag, %s, Sum(Copies) as Total From RecommendedHit GROUP BY %s ORDER BY Total DESC'''
                        % (strLevelToReport, strFullLevels, strFullLevels))
            con.commit()             
            
            wx.MessageBox("Successfully created full report down to '" + strLevelToReport + "'.", caption = 'Task Complete', style=wx.OK)                          
        else:
            wx.MessageBox("No table named RecommendedHit present in SQL file. Try running 'Get Recommended Hits' first.", caption = 'Error', style=wx.ICON_ERROR)        
        
    def GetSingleTaxonomyReport(self):
        '''Outputs a taxonomy report that summarizes the taxon level selected and the count for that level'''
        
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "RecommendedHit" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 
            
            strLevelToReport = cbxlist_TaxonomyLevels[self.options.indexTaxonomySingleOutputLevel]
            
            if strLevelToReport == "Order":
                strLevelToReport = "TaxonOrder"            
            
            cur.execute('''DROP VIEW IF EXISTS SingleTaxonReport_%s''' % (strLevelToReport))
            con.commit()
            cur.execute('''CREATE VIEW SingleTaxonReport_%s AS SELECT %s, Sum(Copies) as Total From RecommendedHit GROUP BY %s ORDER BY %s'''
                        % (strLevelToReport, strLevelToReport, strLevelToReport, strLevelToReport))
            con.commit()
            
            cur.execute('''DROP VIEW IF EXISTS SingleTaxonReportByTag_%s''' % (strLevelToReport))
            con.commit()
            cur.execute('''CREATE VIEW SingleTaxonReportByTag_%s AS SELECT Tag, %s, Sum(Copies) as Total From RecommendedHit GROUP BY %s ORDER BY %s'''
                        % (strLevelToReport, strLevelToReport, strLevelToReport, strLevelToReport))
            con.commit()            
            
            wx.MessageBox("Successfully created single level report for '" + strLevelToReport + "'.", caption = 'Task Complete', style=wx.OK)                          
        else:
            wx.MessageBox("No table named RecommendedHit present in SQL file. Try running 'Get Recommended Hits' first.", caption = 'Error', style=wx.ICON_ERROR)

    def GetRarefactionCurves(self):
        '''This computes the rarefaction curves for the given sample at each taxonomic level (for most compatibility).
        This function uses 100 iterations over the dataset, and the average value from those iterations is saved'''
        
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "RecommendedHit" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 
            
            #Setup table for data input            
            cur.execute('''DROP TABLE IF EXISTS RarefactionCurves''')
            con.commit()
            cur.execute('''CREATE TABLE RarefactionCurves (SampleNumber INTEGER NOT NULL, KingdomCount REAL, PhylumCount REAL,
            ClassCount REAL, OrderCount REAL, FamilyCount REAL, GenusCount REAL, SpeciesCount REAL, PRIMARY KEY (SampleNumber))''')
            con.commit()            

            cur.execute('''SELECT Sum(Copies) FROM RecommendedHit''')
            myquery_sequencedata = cur.fetchall() 
            count_tuple = myquery_sequencedata[0]
            strTotalRecords = str(count_tuple[0])
            
            self.UpdateStatus("Creating Rarefaction Curve Table...")

            cur.execute('''SELECT HitID, Copies FROM RecommendedHit''')
            myquery_sequencedata = cur.fetchall() 

            intHitIDPool = []
            intcounter = 0
            for row in myquery_sequencedata:
                for x in range(int(row[1])):
                    self.UpdateStatus("Adding record #" + str(intcounter) + " of " + strTotalRecords + " to rarefaction sample pool")
                    intHitIDPool.append(int(row[0]))
                    
                    #add default values to all cells in the rarefaction curves table
                    cur.execute('''INSERT INTO RarefactionCurves(SampleNumber, KingdomCount, PhylumCount, ClassCount, OrderCount, FamilyCount, GenusCount, SpeciesCount)
                                VALUES(%s,0,0,0,0,0,0,0)''' % (intcounter))                                
                    intcounter += 1
            con.commit() 
            
            #Create a list for number of unique counts and number of samplings (times not "")       
            intIterations = 100
            strColumns = ["KingdomCount", "PhylumCount", "ClassCount", "OrderCount", "FamilyCount", "GenusCount", "SpeciesCount"]

            for iteration in range(intIterations):
                #during each iteration, check each list in the sample set
                list_setDiscovered = [set(), set(), set(), set(), set(), set(), set()]
                list_sampleAttempts = [0,0,0,0,0,0,0]
                list_IncrementsCarriedThisIteration = [0,0,0,0,0,0,0]
                random.shuffle(intHitIDPool)
            
                #Read each record from the shuffled list, storing data on each taxon level
                for (sampleAttempt, record) in enumerate(intHitIDPool):

                    self.UpdateStatus("Conducting Rarefaction Iteration #" + str(iteration + 1) + " of " + str(intIterations) + ", Sample #" + str(sampleAttempt) + " of " + strTotalRecords)
                    
                    cur.execute('''SELECT Kingdom, Phylum, Class, TaxonOrder, Family, Genus, Species FROM RecommendedHit WHERE HitID = %s''' % (record))
                    myquery_sequencedata = cur.fetchall()                     

                    for (z, column) in enumerate(myquery_sequencedata[0]):
                        #Increment counts for all new records
                        strColumnToUpdate = strColumns[z]

                        list_sampleAttempts[z] += 1
                        
                        if not (column == "") and not (column.lower() in list_setDiscovered[z]):
                            #add new record to set
                            list_setDiscovered[z].add(column.lower())
                            list_IncrementsCarriedThisIteration[z] += 1
                           
                        #increment specific sample's count in database
                        cur.execute('''UPDATE RarefactionCurves SET %s = %s + %s WHERE SampleNumber = %s''' 
                                    % (strColumnToUpdate, strColumnToUpdate, list_IncrementsCarriedThisIteration[z], list_sampleAttempts[z]))
    
                con.commit()        
            #After all iterations are done, divide all counts in the database by the number of iterations performed
            for taxonlevel in strColumns:
                self.UpdateStatus("Conducting Average Rarefaction Across Iterations...")
                cur.execute('''UPDATE RarefactionCurves SET %s = %s/%s''' % (taxonlevel, taxonlevel, intIterations))
                con.commit()                
            
            wx.MessageBox("Successfully created rarefaction curve.", caption = 'Task Complete', style=wx.OK)                          
        else:
            wx.MessageBox("No table named RecommendedHit present in SQL file. Try running 'Get Recommended Hits' first.", caption = 'Error', style=wx.ICON_ERROR)        
 
    def GetShannonIndex(self):
        '''Gets the shannon index by calculating each species proportion, then summing the products
        of the proportion by the natural log of the proportions'''
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
        set_TableNames = set(list_tablenames)
        
        if "RecommendedHit" in set_TableNames:
            database_path = self.sql_filepath
            con = dbapi.connect(database_path)
            cur = con.cursor()
            #Set database output into python to be in regular string format
            con.text_factory = str 
            
            #Setup table for data input            
            cur.execute('''DROP TABLE IF EXISTS ShannonIndex''')
            con.commit()
            cur.execute('''CREATE TABLE ShannonIndex (ShannonID INTEGER NOT NULL, Genus VARCHAR, Species VARCHAR, Count REAL, Proportion REAL,
                        lnProportion REAL, Product REAL, PRIMARY KEY (ShannonID))''')
            con.commit()
            
            cur.execute('''DROP TABLE IF EXISTS ShannonIndexWithUnassigned''')
            con.commit()
            cur.execute('''CREATE TABLE ShannonIndexWithUnassigned (ShannonID INTEGER NOT NULL, Genus VARCHAR, Species VARCHAR, Count REAL,
                        Proportion REAL, lnProportion REAL, Product REAL, PRIMARY KEY (ShannonID))''')
            con.commit()                

            cur.execute('''SELECT Genus, Species, sum(Copies) as Count FROM RecommendedHit GROUP BY Genus, Species ORDER BY Count DESC''')
            myquery_countdata = cur.fetchall() 
            
            #Add all counts to table
            for (x, row) in enumerate(myquery_countdata):
                self.UpdateStatus("Adding species counts #" + str(x) + " to Shannon Index Tables")
                #Everything goes into the Unassigned table
                cur.execute('''INSERT INTO ShannonIndexWithUnassigned (ShannonID, Genus, Species, Count) VALUES (NULL,?,?,?)''',
                            (row[0], row[1], row[2]))
                
                #Only records with both species and genus not = "" go into the normal one
                if (not row[0] == "") and (not row[1] == ""):
                    cur.execute('''INSERT INTO ShannonIndex (ShannonID, Genus, Species, Count) VALUES (NULL,?,?,?)''',
                                (row[0], row[1], row[2]))                    
            
                con.commit()
            
            #Calculate total counts for each table
            intTotalCountShannonIndex = 0
            intTotalCountShannonIndexWithUnassigned = 0
            
            cur.execute('''SELECT sum(Count) FROM ShannonIndex''')
            myquery_shannontotal = cur.fetchall()            
            temp_tuple = myquery_shannontotal[0]
            intTotalCountShannonIndex = temp_tuple[0]
            
            cur.execute('''SELECT sum(Count) FROM ShannonIndexWithUnassigned''')
            myquery_shannontotalALL = cur.fetchall()            
            temp_tuple = myquery_shannontotalALL[0]
            intTotalCountShannonIndexWithUnassigned = temp_tuple[0]
            
            cur.execute('''SELECT ShannonID, Count FROM ShannonIndexWithUnassigned''')
            myquery_indexandcountUnassigned = cur.fetchall()
            intTotalSpecies = len(myquery_indexandcountUnassigned) 
            #Calculate all values for each row and assign them to the right cells
            for (x, row) in enumerate(myquery_indexandcountUnassigned):
                self.UpdateStatus("Calculating proportions for unassigned species #" + str(x) + " of " + str(intTotalSpecies))
                
                dblProportionUnassigned = row[1] / intTotalCountShannonIndexWithUnassigned
                dbllnProportionUnassigned = math.log(dblProportionUnassigned)
                dblProductUnassgined = dblProportionUnassigned * dbllnProportionUnassigned
                
                #Everything goes into the Unassigned table
                cur.execute('''UPDATE ShannonIndexWithUnassigned SET Proportion = %s WHERE ShannonID = %s''' 
                            % (dblProportionUnassigned, row[0]))
                cur.execute('''UPDATE ShannonIndexWithUnassigned SET lnProportion = %s WHERE ShannonID = %s''' 
                            % (dbllnProportionUnassigned, row[0]))                
                cur.execute('''UPDATE ShannonIndexWithUnassigned SET Product = %s WHERE ShannonID = %s''' 
                            % (dblProductUnassgined, row[0])) 
                
                
            cur.execute('''SELECT ShannonID, Count FROM ShannonIndexWithUnassigned''')
            myquery_indexandcount = cur.fetchall()
            intTotalSpecies = len(myquery_indexandcount) 
            #Calculate all values for each row and assign them to the right cells
            for (x, row) in enumerate(myquery_indexandcount):
                self.UpdateStatus("Calculating proportions for assigned species #" + str(x) + " of " + str(intTotalSpecies))              
        
                dblProportion = row[1] / intTotalCountShannonIndex
                dbllnProportion = math.log(dblProportion)
                dblProduct = dblProportion * dbllnProportion
                
                #Everything goes into the Unassigned table
                cur.execute('''UPDATE ShannonIndex SET Proportion = %s WHERE ShannonID = %s''' 
                            % (dblProportion, row[0]))
                cur.execute('''UPDATE ShannonIndex SET lnProportion = %s WHERE ShannonID = %s''' 
                            % (dbllnProportion, row[0]))                
                cur.execute('''UPDATE ShannonIndex SET Product = %s WHERE ShannonID = %s''' 
                            % (dblProduct, row[0]))                     
            
                con.commit()
            
            #Calculate Overall Shannon Index Values, and output to messagebox
            cur.execute('''DROP VIEW IF EXISTS ShannonIndexUnassignedValue''')
            cur.execute('''CREATE VIEW ShannonIndexUnassignedValue AS SELECT sum(Product)*-1 as ShannonIndexUnassignedValue FROM ShannonIndexWithUnassigned''')
            cur.execute('''SELECT sum(Product)*-1 FROM ShannonIndexWithUnassigned''')
            myquery_SIunassigned = cur.fetchall()            
            temp_tuple = myquery_SIunassigned[0]
            dblSIUnassinged = temp_tuple[0]                        
            
            cur.execute('''DROP VIEW IF EXISTS ShannonIndexValue''')
            cur.execute('''CREATE VIEW ShannonIndexValue AS SELECT sum(Product)*-1 as ShannonIndexValue FROM ShannonIndex''')
            cur.execute('''SELECT sum(Product)*-1 FROM ShannonIndex''')
            myquery_SI = cur.fetchall()           
            temp_tuple = myquery_SI[0]
            dblSI = temp_tuple[0]             

            message = '''Successfully calculated Shannon Index. The value for your sample is:
            
INCLUDING records with unnssigned genus or species = %s

EXCLUDING records with unassigned genus or species = %s''' % (dblSIUnassinged, dblSI)
            wx.MessageBox(message, caption = 'Task Complete', style=wx.OK)                          
        else:
            wx.MessageBox("No table named RecommendedHit present in SQL file. Try running 'Get Recommended Hits' first.", caption = 'Error', style=wx.ICON_ERROR)        
        
    def ExportAllFiles(self):
        '''This function goes through all the tables and views in the SQL file provided, and outputs them to Excel csv'''
        
        list_tablenames = self.GetListOfTablesAndViews(self.sql_filepath)
            
        for strName in list_tablenames:
            self.OutputExcelFile(self.sql_filepath, strName, boolExportAll=True)
            
        wx.MessageBox(message = "All tables and views saved as Excel files at:\n" + os.path.dirname(self.sql_filepath), caption = "Save Successful")  
    
    def CombineSQLFiles(self, window):
        '''Function that combines the input SQL files, generating an output SQL database
        with only the unique entries in the final version (removes duplicate rows). This
        is useful for any analyses that were split up for faster processing, and allows
        for some overlap in the ranges to occur without issue (provided the settings were
        the same'''
        
        #Get the files to combine
        window.dlg = wx.FileDialog(window,
                                 message="Select Files to combine",
                                 wildcard="SQL Files (*.sqlite)|*.sqlite",
                                 style=wx.FD_MULTIPLE)
        window.dlg.ShowModal()
    
        if not (window.dlg.Paths is None):
            if len(window.dlg.Paths) <= 1:
                wx.MessageBox("Combine requires at LEAST 2 SQL files to combine","File Count Error", wx.ok)
            else:
                sql_filepaths = window.dlg.Paths
                sql_filenames = window.dlg.Filenames
                
                wx.MessageBox("Please select a NEW name for the combined file", "Need Output Filename")
            
                #Get the filename to save new Database as
                window.dlg = wx.FileDialog(window,
                                 message="Save Combined Database As...",
                                 wildcard="SQL Files (*.sqlite)|*.sqlite",
                                 style=wx.FD_SAVE)
                window.dlg.ShowModal()
                
                if window.dlg.Paths:
                    sql_savepath = window.dlg.Paths[0]
                    
                    #Make sure file extension is correct
                    if not sql_savepath.endswith(".sqlite"):
                        sql_savepath = sql_savepath + ".sqlite"
                    sql_savename = sql_savepath.split("\\")[-1]
                
                    #Get list of all tablenames in each file, using set so unique names only are collected
                    set_tablenames = set()
                    for source_database in sql_filepaths:
                        database_path = source_database
                        con = dbapi.connect(database_path)
                        cur = con.cursor()
                        #Set database output into python to be in regular string format
                        con.text_factory = str 
                        
                        cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
                        temp_tablenames = cur.fetchall()
                        
                        #Bugfix - table[0] needed as it is a list of list format with only 1 record
                        for table in temp_tablenames:
                            set_tablenames.add(table[0])
                    
                    #Convert table set to list
                    list_tablenames = list(set_tablenames)
                    
                    con_save = dbapi.connect(sql_savepath)
                    cur_save = con_save.cursor()
                    con_save.text_factory = str
                    
                    #ERROR HANDLING = PRAGMA statements fix issues related to very very large databases
                    cur_save.execute('''PRAGMA synchronous = OFF;''')
                    cur_save.execute('''PRAGMA journal_mode = MEMORY;''')
                    cur_save.execute('''PRAGMA cache_size=500000;''')
                    cur_save.execute('''PRAGMA main.page_size = 4096;''')
                    con_save.commit()
                    cur.execute('''PRAGMA synchronous = OFF;''')
                    cur.execute('''PRAGMA journal_mode = MEMORY;''')
                    cur.execute('''PRAGMA cache_size=500000;''')
                    cur.execute('''PRAGMA main.page_size = 4096;''')
                    con.commit()                
                    
                    #Go through each table and combine from all databases
                    for table in list_tablenames:
                        self.UpdateStatus("Combining all instances of Table '%s'" % (table))
                        for source_database in sql_filepaths:
                            source_name = source_database.split("\\")[-1]
                            database_path = source_database
                            con = dbapi.connect(database_path)
                            cur = con.cursor()
                            con.text_factory = str 
                            
                            cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
                            myquery_tables = cur.fetchall()
                            set_toMergeTables = set()
                            for value in myquery_tables:
                                set_toMergeTables.add(value[0])
                            
                            #Located here so it updates set on each new source database
                            cur_save.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
                            myquery_savetables = cur_save.fetchall()
                            set_savetables = set()
                            for value in myquery_savetables:
                                set_savetables.add(value[0])                        
    
                            #Check to see if table exists in source database, otherwise skip to next source database
                            if table in set_toMergeTables:
                                #If table is new to the save database, create it first
                                if not table in set_savetables:
                                    #Get SQL creation code for table from original database
                                    cur.execute('''SELECT sql FROM sqlite_master WHERE tbl_name ="%s";''' % (table))
                                    myquery_creationSQLcode = cur.fetchall()
                                    
                                    #This should only run once, but handles error of blank first row
                                    for code in myquery_creationSQLcode:
                                        creationSQLcode = code[0]
                                        
                                    #Create New Table in Save Database with Creation code from toMerge
                                    cur_save.execute(creationSQLcode)
                                    con_save.commit()                            
    
                                #Get list of columns in table
                                #Don't get the ID column, as it should autoincrement if left out as NULL
                                #NOTE: this assumes first row is key row, but that is just good formating
                                cur.execute('''PRAGMA table_info(%s);''' % (table))
                                myquery_tablecolumns = cur.fetchall()
                                
                                allcolumnstring = ""
                                for column in myquery_tablecolumns:
                                    if not allcolumnstring == "":
                                        allcolumnstring = allcolumnstring + ", "
                                    allcolumnstring = allcolumnstring + column[1]
    
                                insertcolumnstring = "NULL"
                                for column in myquery_tablecolumns[1:]:
                                    if not insertcolumnstring == "":
                                        insertcolumnstring = insertcolumnstring + ", "
                                    insertcolumnstring = insertcolumnstring + column[1]
                                    
                                temp_column = myquery_tablecolumns[0]
                                IDcolumn = temp_column[1]
                                
                                #Insert data into table with the "INSERT INTO SELECT" method
                                cur_save.execute('''ATTACH DATABASE "%s" as toMerge;''' % (source_database))
                                con_save.commit()
                                
                                index = 1
                                blocksize = 1000
                                while True:
                                    #Get row from source table
                                    cur.execute('''SELECT * FROM %s WHERE %s = %s''' % (table, IDcolumn, index))
                                    myquery_onerow = cur.fetchone()
                                    
                                    if myquery_onerow is None:
                                        break
                                    else:
                                        #Remove first entry of list, so it can be replaced by null
                                        list_onerow = myquery_onerow[1:]
                                        
                                        #Insert row into save table
                                        string_sql = 'INSERT INTO ' + table + ' (' + allcolumnstring + ') VALUES (NULL ' + ",?"*len(list_onerow) + ');'
                                        cur_save.execute(string_sql, list_onerow)
                        
                                        if index%blocksize == 0:
                                            #Commit Block of inserts
                                            con_save.commit()
                                            
                                        index += 1
                                                                
                                #Commit the last block
                                con_save.commit()
                                    
                                cur_save.execute('''DETACH DATABASE toMerge;''')
                                con_save.commit()
                            con.close()
                                
                    con_save.close()
        
                self.SQLRemoveDuplicates(window, sql_savepath)
                
        window.dlg.Destroy()
            
    def SQLRemoveDuplicates(self, window, sql_filepath):
        '''Goes through all tables in the supplied sql_filepath file, and removes all duplicate rows. This could be time-intensive'''
       
        self.UpdateStatus("Removing Duplicate Records")
        database_path = sql_filepath
        con = dbapi.connect(database_path)
        cur = con.cursor()
        #Set database output into python to be in regular string format
        con.text_factory = str 
        
        cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
        temp_tablenames = cur.fetchall()
        
        #Bugfix - table[0] needed as it is a list of list format with only 1 record
        set_tablenames = set()
        for table in temp_tablenames:
            set_tablenames.add(table[0])  
    
        for table in set_tablenames:
            self.UpdateStatus("Removing Duplicate Records in Table '%s'" % (table))
            tabletemp = table + "_temp"
            
            #Get SQL creation code for table
            cur.execute('''SELECT sql FROM sqlite_master WHERE tbl_name ="%s";''' % (table))
            myquery_creationSQLcode = cur.fetchall()
            
            #This should only run once, but handles error of blank first row
            for code in myquery_creationSQLcode:
                creationSQLcode = code[0]
                
            #Rename current table to _temp
            cur.execute('''ALTER TABLE %s RENAME TO %s''' % (table, tabletemp))
            con.commit()   
            
            #Create fresh table with the original name
            cur.execute(creationSQLcode)
            con.commit()                            
    
            #Get list of columns in table
            #Don't get the ID column, as it should autoincrement if left out as NULL
            #NOTE: this assumes first row is key row, but that is just good formating
            cur.execute('''PRAGMA table_info(%s);''' % (tabletemp))
            myquery_tablecolumns = cur.fetchall()
            
            allcolumnstring = ""
            for column in myquery_tablecolumns:
                if not allcolumnstring == "":
                    allcolumnstring = allcolumnstring + ", "
                allcolumnstring = allcolumnstring + column[1]
    
            insertcolumnstring = ""
            sortcolumnstring = ""
            boolSortStringFound = False
            for column in myquery_tablecolumns[1:]:
                if not boolSortStringFound:
                    sortcolumnstring = column[1]
                    boolSortStringFound = True
                    
                if not insertcolumnstring == "":
                    insertcolumnstring = insertcolumnstring + ", "
                insertcolumnstring = insertcolumnstring + column[1]
    
            #Save All distinct rows into the new table
            cur.execute('''INSERT INTO %s(%s) SELECT DISTINCT NULL, %s FROM %s ORDER BY %s''' % (table, allcolumnstring, insertcolumnstring, tabletemp, sortcolumnstring))
            con.commit()
            
            cur.execute('''DROP TABLE IF EXISTS %s''' % (tabletemp))
    
        wx.MessageBox("Database files succesfully combined, and all duplicates removed!","Task Complete", style=wx.OK)

    def GetListOfTablesAndViews(self, sql_filepath):
        '''This function creates a list of all table and view names from the provided SQL file'''
        #Get list of tables in SQL file
        database_path = self.sql_filepath
        con = dbapi.connect(database_path)
        cur = con.cursor()
        con.text_factory = str 
        
        cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
        temp_tablenames = cur.fetchall()
        
        #Bugfix - table[0] needed as it is a list of list format with only 1 record
        list_tablenames = []
        for table in temp_tablenames:
            list_tablenames.append(table[0])
            
        cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="view";''')
        temp_viewnames = cur.fetchall()
                
        #Bugfix - view[0] needed as it is a list of list format with only 1 record
        for view in temp_viewnames:
            list_tablenames.append(view[0])
            
        return list_tablenames
    
    def GetSQLFilePath(self):
        '''This gets the SQL file by prompting with a filebrowser'''
        bool_file_selected = False
    
        #Open File Browser
        try:
            window = wx.Window(parent=self.parent)
            
            window.dlg = wx.FileDialog(window,
                                       message="Select SQLite File(s) to analyze",
                                       wildcard="SQLite Files(.sqlite)|*.sqlite",
                                       style=wx.FD_DEFAULT_STYLE)
            window.dlg.ShowModal()
            
            if window.dlg.Paths:
                paths = window.dlg.Paths
                self.sql_filepath = os.path.dirname(paths[0])
                window.dlg.Destroy()
                bool_file_selected = True
            else:
                bool_file_selected = False
                window.dlg.Destroy()
        except:
            bool_file_selected = False 

        return bool_file_selected
        
        
    def OutputExcelFile(self, sql_filepath, tablename, boolExportAll=False):
        '''Saves the given table as an Excel file to the folder that contains the SQL file'''
        excel_filepath = os.path.dirname(sql_filepath) + "\\" + self.options.strProjectName + "_" + tablename + ".csv"
        
        database_path = sql_filepath
        con = dbapi.connect(database_path)
        cur = con.cursor()
        #Set database output into python to be in regular string format
        con.text_factory = str 
            
        #Get list of columns in the table
        cur.execute('''PRAGMA table_info(%s);''' % (tablename))
        myquery_tablecolumns = cur.fetchall()
        columnstring = []
        for column in myquery_tablecolumns[:]:
            columnstring.append(column[1])#column[1] is columnname index   
        
        #Get the table data    
        cur.execute('''SELECT * FROM %s''' % (tablename))  
        myquery_Alltabledata = cur.fetchall()        

        try:
            outputfile = open(excel_filepath, 'wb')
            writer = csv.writer(outputfile)
            writer.writerow(columnstring)
            writer.writerows(myquery_Alltabledata)
            outputfile.close()
            
            if (not boolExportAll):
                wx.MessageBox(message = tablename + " saved as Excel file at:\n" + excel_filepath, caption = "Save Successful")            
        except:
            wx.MessageBox(message = tablename + " was NOT saved as Excel file\n" + "Cannot access save file location: " + excel_filepath + ".\n Please close the file and try again.",
                          caption = "Error: Output File In Use",
                          style=wx.ICON_ERROR)
        
        return excel_filepath
        
    def UpdateStatus(self, message):
        '''Updates the status bar with the supplied message.'''
        pub.sendMessage("object.added", message=message)
        
        
##Read from the .fna file, create .fasta file
def CreateFASTA_FromFNA(fna_filepaths, project_name, window):
    '''Creates a FASTA format file out of the 454 Output .FNA file. Can accept
    multiple .fna files at once, and will combined them into 1 FASTA file.'''
    
    FolderForOutput = os.path.dirname(fna_filepaths[0])
    output_name = os.path.dirname(fna_filepaths[0]) + "\\" + project_name + ".fasta"

    for filepath in fna_filepaths:
        input_file = open(filepath, 'rU')
        entire_file = input_file.read()
        input_file.close()
        
        #Split file on ">", which signals new sequence records
        sequences = entire_file.split(">")
    
        #Split each sequence into the components, keep only the name and sequence
        #[Name, Rank, x, y, length, sequences...] and write to FASTA file
        output_file = open(output_name, 'wb')
        sequence_number = 0
        for sequence in sequences[1:]:
            sequence_number += 1
            components = sequence.split(None)
            name = components[0]
            sequencelines = components[5:]
            
            Sequence454 = Seq454(name, sequencelines)
                    
            output_file.write(Sequence454.FASTA())
            
            if sequence_number%25 == 0:
                window.UpdateStatus("Creating FASTA Files... Records: %s" % (sequence_number))
        output_file.close()
        
    return output_name

def CreateFASTA_FromFASTA(fna_filepaths, project_name, window):
    '''Creates a FASTA format file out of the .FASTA, .FA, or .TXT file. Can accept
    multiple files at once, and will combined them into 1 FASTA file.'''
    
    output_name = os.path.dirname(fna_filepaths[0]) + "\\" + project_name + ".fasta"

    for filepath in fna_filepaths:
        input_file = open(filepath, 'rU')
        entire_file = input_file.read()
        input_file.close()
        
        #Split file on ">", which signals new sequence records
        sequences = entire_file.split(">")
    
        #Split each sequence into the components, keep only the name and sequence
        #[Name, Rank, x, y, length, sequences...] and write to FASTA file
        output_file = open(output_name, 'wb')
        sequence_number = 0
        for sequence in sequences[1:]:
            sequence_number += 1
            components = sequence.split("\n")
            name = components[0]
            sequencelines = components[1:]
            
            Sequence454 = Seq454(name, sequencelines)
                    
            output_file.write(Sequence454.FASTA())
            if sequence_number%25 == 0:
                window.UpdateStatus("Creating FASTA Files... Records: %s" % (sequence_number))
        output_file.close()
    
    return output_name
        
def ParseSQLite_IntoSeq454(filepath, window, range_lower = -1, range_upper = -1):
    '''Returns a list of Seq454 Objects containing the sequence data from the 
    supplied SQLite file. The range values are for use with the only select sequences
    from between the values, to decrease memory usage.'''
    list_Seq454 = []
        
    con = dbapi.connect(filepath)
    cur = con.cursor()
    
    #Set database output into python to be in regular string format
    con.text_factory = str 
    
    if range_lower == -1 and range_upper == -1:
        #Get all the SequenceData Table from the file - only sqlite files made with this program work
        cur.execute('''SELECT SeqID, SeqName, Copies, Tag, Sequence
                       FROM SequenceData 
                       ORDER BY SeqID''')
        myquery_sequencedata = cur.fetchall()        
        
                       
    else:
        #Get range of data from the file
        cur.execute('''SELECT SeqID, SeqName, Copies, Tag, Sequence
                       FROM SequenceData 
                       WHERE SeqID > %s and SeqID <= %s
                       ORDER BY SeqID''' % (range_lower, range_upper))
        myquery_sequencedata = cur.fetchall()
        

    int_count = 0
    for datarow in myquery_sequencedata:
        #Add all seqeuences to list_Seq454
        if not datarow[4] == "":
            int_count += 1
            temp_Seq454 = Seq454(datarow[1],datarow[4],datarow[2])
            temp_Seq454.tag = datarow[3]

            list_Seq454.append(temp_Seq454)
            
            #if int_count%25 == 0:
            #    window.UpdateStatus("Creating 454 Objects... Records: %s" % (int_count))  
            
    return list_Seq454

def ParseFASTA_IntoSeq454(filepath, window):
    '''Returns a list of Seq454 Objects containing the sequence data from the 
    supplied FASTA file. If the name of a sequence in the FASTA file contains
    a count value in parentheses, sets that Object's count to that number,
    otherwise defaults to 1.'''
    list_Seq454 = []
    
    input_file = open(filepath, 'rU')
    entire_file = input_file.read()
    input_file.close()
        
    #Split file on ">", which signals new sequence records
    sequences = entire_file.split(">")
    
    #Split each sequence into the components, and check for the count value
    #if it is present
    int_count = 0
    for sequence in sequences[1:]:
        int_count += 1
        components = sequence.split("\n")
        name = components[0]
        sequencelines = components[1:]
        count = 1
                
        list_Seq454.append(Seq454(name, sequencelines, count))
        
        if int_count%25 == 0:
            window.UpdateStatus("Creating 454 Objects... Records: %s" % (int_count))
 
    return list_Seq454
    
    
##Split .fasta file into files by sequence tags
def EditSeq454Tags(list_454, project_name, window):
    '''Takes the full 454 list and determines the tags for each sequence.'''

    int_count = 0
    for seq454 in list_454:
        int_count += 1
        selected_tag = seq454.determine_tag()
        seq454.tag = selected_tag

        if int_count%25 == 0:
            window.UpdateStatus("Editing Sequence Tags... Record: %s" % (int_count))
    
##Combine exactly same sequences, rename using the IDTag label
def CombineAndRenameIdenticalSequences(list_454, IDTag, window):
    list_combined_454 = []
    int_counter = 0
    record_count = 0
    
    #Bug Fix - using dictionaries to reduce search and combine times
    #to a 2 passes (1 pass creating dictionary, 1 pass searching for presense) and not n^2    
    dict_seq_counts = dict() #Key = DNA sequence, Value = count of reps
    dict_seq_newID = dict() #Key = DNA sequence, Value = new sequence ID
    
    for seq454 in list_454:
        record_count += 1
        #Bug Fix - skip all empty sequences, which sometimes result from string.split(">")
        if not seq454.seq == "":            
            #Use get statement to creat new dict element if needed, or increment if existing
            dict_seq_counts[seq454.seq] = dict_seq_counts.get(seq454.seq, 0) + 1
            
            if dict_seq_counts[seq454.seq] == 1: #This is a new record
                #no match was found, add new record to combined list
                int_counter += 1
                temp_name = IDTag + str(int_counter)
                temp_combined454 = Seq454(temp_name, seq454.seq)
                temp_combined454.tag = seq454.tag
                temp_combined454.original_name = seq454.original_name
                list_combined_454.append(temp_combined454)
                
                #add to the ID list for combining
                dict_seq_newID[seq454.seq] = temp_name
                
                if record_count%25 == 0:
                    window.UpdateStatus("Combining Identical Sequences... %s records into %s" % (record_count, int_counter))                
                    
            elif dict_seq_counts[seq454.seq] > 1: #This record already exists
                temp_name = dict_seq_newID[seq454.seq]
                
                #get record number (i.e. position in the combined list) from the temp_name
                splitname = temp_name.split(IDTag)
                ID_index = int(splitname[1]) - 1 #(IDTag, ID_Number), where ID_number is 1 larger than the index in the list
                
                list_combined_454[ID_index].increment()
                list_combined_454[ID_index].original_name += ", " + seq454.name
                
                if record_count%25 == 0:
                    window.UpdateStatus("Combining Identical Sequences... %s records into %s" % (record_count, int_counter))                
         
            else: #there is some count assignment error 
                wx.MessageBox("Invalid count of sequence records in combining function", "Error") 

    return list_combined_454

def SaveCombined454Files(list_combined_454, project_name, window, fna_filepaths):
    '''Saves the new renamed and combined lists into files. These will be mostly
    in FASTA format, but an additional part of the name will be "<name>(count)"
    to keep track of how many sequences were combined into one. This are just
    extra output files for researcher records. This also saves the first tables
    of the SQLite database, so pre-processing doesn't need to be repeated.'''
    
    database_path = os.path.dirname(fna_filepaths[0]) + "\\" + project_name + "_Analysis.sqlite"
    con = dbapi.connect(database_path)
    cur = con.cursor()
    
    #Set database output into python to be in regular string format
    con.text_factory = str
   
    #Drop tables if they exist since the combine step only saves once and not if source is sql
    cur.execute('''DROP TABLE IF EXISTS OldNewNames''')
    con.commit() 
    cur.execute('''DROP TABLE IF EXISTS SequenceData_All''')
    con.commit()     
            
    #Create Table - Old and New Names
    cur.execute('''CREATE TABLE IF NOT EXISTS OldNewNames
    (NameID INTEGER NOT NULL, OldName VARCHAR, NewName VARCHAR NOT NULL,
    PRIMARY KEY (NameID))''')
    con.commit()    
  
    #Create Table - SequenceData
    cur.execute('''CREATE TABLE IF NOT EXISTS SequenceData_All
    (SeqID INTEGER NOT NULL, SeqName VARCHAR, Copies INTEGER,
    Tag VARCHAR, Sequence VARCHAR, CompleteSequence VARCHAR,
    PRIMARY KEY (SeqID))''')
    con.commit()    
        
    #Iterate Through all Sequences
    int_count = 0
    for combined_454 in list_combined_454:
        int_count += 1
        window.UpdateStatus("Writing to SQL File... Records: %s" % (int_count))
               
        #Populate Original Names Table
        original_names = combined_454.original_name.split(", ")
        for original in original_names:
            try:
                cur.execute('''INSERT INTO OldNewNames VALUES(?,?,?)''', (None, original, combined_454.name))
            except:
                #Bug Fix: sleeping allows delay between entries which sometimes occur too quickly to update correctly on large files
                time.sleep(0.5)
                cur.execute('''INSERT INTO OldNewNames VALUES(?,?,?)''', (None, original, combined_454.name))                
        con.commit()

        #Populate Sequence Data Table
        tag = combined_454.tag
        sequencewithouttag = combined_454.seq
        completesequence = tag + sequencewithouttag
        
            
        try:
            cur.execute('''INSERT INTO SequenceData_All VALUES(?,?,?,?,?,?)''', (None, combined_454.name, combined_454.count, tag, sequencewithouttag, completesequence))
        except:
            #Bug Fix: sleeping allows delay between entries which sometimes occur too quickly to update correctly on large files
            time.sleep(0.5)
            cur.execute('''INSERT INTO SequenceData_All VALUES(?,?,?,?,?,?)''', (None, combined_454.name, combined_454.count, tag, sequencewithouttag, completesequence))
                    
        con.commit()
        
    #returns the location to the database, rather than the files, to save RAM usage    
    return database_path 
        


##Defines The Blast Analysis Master Thread
class ConductBLASTAnalysis(wx.Frame):
    '''Class for the Blast Analysis Frame'''
    def __init__(self, parent, notify_window, sql_filepath, options):
        '''Calls the BLAST analysis program through either multiprocessing or 
        series calls. Updates the statusbar with progress. Returns list of analyzed
        Seq454 objects.'''
        
        wx.Frame.__init__(self,  notify_window)           
        
        #Form Variables        
        self._notify_window = notify_window
        self.parent = parent
        self.options = options
        self.sql_filepath = sql_filepath
        self.message = ""
    
        con = dbapi.connect(self.sql_filepath)
        cur = con.cursor()  
        
        #Get list of tablenames in database
        cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
        temp_tablenames = cur.fetchall()
        
        #Bugfix - table[0] needed as it is a list of list format with only 1 record
        list_tablenames = []
        for table in temp_tablenames:
            list_tablenames.append(table[0])
        
        self.total_sequences = 0
        self.total_sequences = FindSkippedSequences(self.options, self._notify_window, self.sql_filepath)
        
        #Boolean variarible indicating that last block is being processed
        self.bool_IsLastBlock = False
        
        #Variable for how many sequences per cpu to use, theortically higher = less downtime
        self.seq_per_cpu = 3
        
        #NOTE: this should be max of 2 to prevent URL lockout by blast server
        self.seq_block_size = self.options.intBLASTSimultaneousThreads
        
        #Set up event handler for worker thread results
        EVT_BLOCKRESULT(self, self.OnBlockResult)
        
        #Indicate that we don't have a worker thread yet (this variable will prevent
        #the program from running multiple work threads at the same time
        self.worker = None
        
        #Initilize the starting and ending position and current block variables
        self.start_number = 0
        self.end_number = self.total_sequences
        self.current_seq_block = 0
        
        if self.options.boolMultipleComputerAnalysis:
            if  self.options.intMultipleComputerTotalNumber > 1:
                #Only runs if more than one total computer, if only one, current start and end positions are fine
                intSequencesPerComputerFloor = math.floor(self.total_sequences/self.options.intMultipleComputerTotalNumber)
                
                self.start_number = (self.options.intMultipleComputerIndex - 1) * intSequencesPerComputerFloor
                
                if self.options.intMultipleComputerIndex == self.options.intMultipleComputerTotalNumber:
                    #If last computer, pick up the remainder sequences from the issue of dividing uneven sequences counts
                    self.end_number = self.total_sequences
                else:
                    self.end_number = (self.options.intMultipleComputerIndex) * intSequencesPerComputerFloor            

    def start(self):
        '''Starts running the block of sequences in the frame when commanded by parent'''
        #Start the worker for the first block
        if not self.worker:
            self.message = "Starting First Sequence Block..."
            pub.sendMessage("object.added", message=self.message)
            self.StartBlockofSequences()
                
    def StartBlockofSequences(self):
        '''Determine the current block of sequences to run through blast, then trigger the thread'''
        #Trigger the worker thread unless its already busy
        if not self.worker:
            self.worker = True
            #Set up the list for the sequences and options to process
            list_Seq454 = []
            
            #Determine index of sequence at start of current block
            current_sequence_index = (self.current_seq_block * self.seq_block_size) + self.start_number
            
            #Bug Fix: Handles the case of "Inappropriate argument value (of correct type)
            #Exceptions.ValueError when current_sequence index == 0
            if current_sequence_index <= 0: current_sequence_index = 0            
            
            #Make sure the end of block position doesn't go past the end of the sequences
            if current_sequence_index + self.seq_block_size <= self.end_number:
                int_end_of_block_position = current_sequence_index + self.seq_block_size
            else:
                int_end_of_block_position = self.end_number
            
            #See if it is the last block to process
            if int_end_of_block_position == self.end_number:
                self.bool_IsLastBlock = True
                                
            #Run analysis for each block of sequences, allowing 10 attempts       
            average_time = 0
            
            #Get the temporary 454 object
            list_Seq454 = ParseSQLite_IntoSeq454(self.sql_filepath, self,
                                                 current_sequence_index, int_end_of_block_position)
            
            try:
                start_time = time.time()
            
                dict_results = dict()
                list_Seq454_processed = []
                threads = []
                q = Queue()
                o = Queue()
            
                #Create thread pool
                for i, seq_index in enumerate(range(current_sequence_index, int_end_of_block_position)):
                    #WAIT 3 SECONDS - NEVER REMOVE THIS AS IT AVOIDS BLAST LOCKOUT
                    time.sleep(3)
                    
                    arguments = (list_Seq454[i], self.options, q, o, i)
                    t = threading.Thread(target=BLASTAnalysis, args=arguments)
                    t.start()
                    threads.append(t)
                
                #Wait for all threads to complete
                for t in threads:
                    t.join()
                    
                #Resort by the ordered pool
                for i, seq_index in enumerate(range(current_sequence_index, int_end_of_block_position)):
                    for entry in o.queue:
                        if entry == i:
                            list_Seq454_processed.append(q.queue[entry])
                            break
                            
                last_time = time.time()
                                
                #Calculate Time Elapsed Per Sequence
                if not len(list_Seq454_processed) == 0:
                    average_time = (last_time - start_time)/len(list_Seq454_processed)
                else:
                    #Bug Handling in case of division by 0
                    average_time = (last_time - start_time)
                 
                #Append Time Per Sequence Data to Benchmarking File
                benchmark_filename = (os.path.dirname(os.path.realpath(__file__)) + "\Benchmarking.txt")     
                benchmark_file = open(benchmark_filename, 'a')
                benchmark_file.write("Average Time For Block:," + str(average_time) + ",seconds,\n")
                benchmark_file.close()    
                
                #Outputs results of block, if no error occurs (which is indicated by the name of the top hit being "Error Occurred"
                #As this is the only way to pass the error out of the daughter thread without having the program kill itself completely
                #If error occurs, just skips the output step so no bad data is saved, and it should be caught on the second pass when the
                #connection issue is unlikely to happen again (most of the errors are just communication problems)
                bool_Error_Occurred = False
                
                for result in list_Seq454_processed:
                    if result.tophits[0].name == "Error Occurred":
                        bool_Error_Occurred = True
                
                if bool_Error_Occurred == False:
                    OutputResults(list_Seq454_processed, self.options.strProjectName, self.options.fna_filepaths, self.options.boolBLASTSaveAllHits, self.options.boolBlastSaveUnculturedClones)
                
            except: # catch *all* exceptions
                e = sys.exc_info()[0]
                
                #For now just keeps going past other Errors (usually some issue with connection)
                #Not optimal, but this prevents any issue with saving
                errorlog_filename = (os.path.dirname(os.path.realpath(__file__)) + "\Error Log.txt")     
                errorlog_file = open(errorlog_filename, 'a')
                errorlog_file.write(str(datetime.datetime.now()) + "\n")
                errorlog_file.writelines(str(e))
                errorlog_file.close()
                time.sleep(5)
                pass
                

            wx.PostEvent(self, ResultEvent(average_time, EVT_BLOCKRESULT_ID))             
        else:
            #Wait before retrying triggering worker thread
            time.sleep(10)
            self.StartBlockofSequences()
             
                  
    def OnBlockResult(self, event):
        '''Show Result status'''
        if event.data:
            average_time = round(float(event.data),2)
            #Update Status with Progress (data from event is average time/seq)
            current_sequence = (self.current_seq_block + 1) * self.seq_block_size
            sequences_remaining = self.total_sequences - current_sequence
            remaining_estimate = average_time * sequences_remaining
            
            hours, remainder = divmod(remaining_estimate, 3600)
            minutes, seconds = divmod(remainder, 60)             
            
            message = ""
            
            if self.bool_IsLastBlock == False:
                self.worker = None
                self.current_seq_block += 1
                self.StartBlockofSequences()
                self.message = "BLASTed %s of %s sequences. Average Time: %s Remaining: %s:%s:%s" % (current_sequence, self.total_sequences, average_time, hours, minutes, seconds)
            else:
                self.message = "Finished BLAST Processing of %s sequences. Average Time: %s sec." % (self.total_sequences, average_time)

            pub.sendMessage("object.added", message=self.message)

##Run Blast analysis
def BLASTAnalysis(Seq454Object, options, queue, order, i):
    '''Runs the community blast analysis with input Seq454 Object and the options
    specified by the input options object. Processes result and modifies the
    Seq454Object. Returns the Seq454Object (for overwriting old record when
    running in parallel).
    NOTE: Inputs to this function HAVE to be as a tuple so that when using
    multiprocessing only 1 input iterable list is sent to the pool.map function.'''
    
    fasta_string = Seq454Object.FASTA()
    
    try:
        result_handle = NCBIWWW.qblast(program="blastn",
                                           database="nr",
                                           sequence=fasta_string,
                                           descriptions=cbxlist_MaxResults[options.indexBLASTMaxResultsPerSequence],
                                           alignments=cbxlist_MaxResults[options.indexBLASTMaxResultsPerSequence],
                                           hitlist_size=cbxlist_MaxResults[options.indexBLASTMaxResultsPerSequence])

    
            ##The following code is vital to handling errors that occur when the returned value stored to
            ##result_handle is not a complete .xml file. This occurs when the number of records in the BLAST
            ##file is somewhere above 10,000 and is possibly due to some delay cutoff on how long the function
            ##waits before calling the fine done. If this is not included, it generates a "missing token" error
            ##in the xml parser function part (not exact name) of NCBIXML.read(). To fix this, I simply check
            ##the returned file for the correct end of file code, and if it is absent, chew back the last partial
            ##record to the </Hit> of the previous record, then concatenate a functional (though incorrect in 
            ##the fine details specific to that query) end of file string to the end, allowing it to get past
            ##this error. Rewriting the BioPython function responsible for the error is necessary to truly fix
            ##this issue, but this solution preserves most of the data from the query in functional form.
            
        contents = result_handle.getvalue()
        
        #Only do the cutting and appending if the end of file "</BlastOutput>" is missing
        if contents.rfind("</BlastOutput>") == -1:
            contents_split = contents[:contents.rfind("</Hit>")]
            fix_string = '''</Hit>
            </Iteration_hits>
              <Iteration_stat>
                <Statistics>
                  <Statistics_db-num>17199598</Statistics_db-num>
                  <Statistics_db-len>1171634340</Statistics_db-len>
                  <Statistics_hsp-len>0</Statistics_hsp-len>
                  <Statistics_eff-space>0</Statistics_eff-space>
                  <Statistics_kappa>0.41</Statistics_kappa>
                  <Statistics_lambda>0.625</Statistics_lambda>
                  <Statistics_entropy>0.78</Statistics_entropy>
                </Statistics>
              </Iteration_stat>
            </Iteration>
            </BlastOutput_iterations>
            </BlastOutput>
            '''
            #reassign to the same variable so whether it takes this appending route or not it works
            contents = contents_split + fix_string
     
        #Restore contents object to original stringIO format
        result_handle = StringIO.StringIO(contents)
        
        blast_record = NCBIXML.read(result_handle)
        max_score = 0
        for alignment in blast_record.alignments:
            for hsp in alignment.hsps:
                if (hsp.expect < options.strBLASTEvalue) and (hsp.identities > options.intBLASTMinimumAlignmentLength):
                    #Fill in correct lists with objects (cultured/uncultured)
                    if "uncultured" in str(alignment.title.lower()):
                        #Get Result Object
                        blast_result = BLASTResult(input_query_length=len(Seq454Object.seq),
                                                   asc = alignment.accession,
                                                   name = alignment.hit_def,
                                                   score = hsp.score,
                                                   evalue = hsp.expect,
                                                   identities = hsp.identities,
                                                   subject_as_string = str(hsp))
                        Seq454Object.unculturedhits.append(blast_result)
                    else:
                        #Get Result Object
                        blast_result = BLASTResult(input_query_length=len(Seq454Object.seq),
                                                   asc = alignment.accession ,
                                                   name = alignment.hit_def,
                                                   score = hsp.score,
                                                   evalue = hsp.expect,
                                                   identities = hsp.identities,
                                                   subject_as_string = str(hsp))
                        Seq454Object.allhits.append(blast_result)
                        
                        #Add to tophits list if it qualifies
                        if blast_result.score > max_score:
                            max_score = blast_result.score
                            Seq454Object.tophits = []
                            Seq454Object.tophits.append(blast_result)
                        elif blast_result.score == max_score:
                            Seq454Object.tophits.append(blast_result)
    
        no_significant_hit_result = BLASTResult(input_query_length=1,
                                                asc = "",
                                                name = "No Significant Hits Found",
                                                score = 0,
                                                evalue = 1e-0,
                                                identities = 0,
                                                subject_as_string = "")
        
        #Fill in any empty lists of hits with no significant hit found message
        if not Seq454Object.tophits:
            Seq454Object.tophits.append(no_significant_hit_result)
        if not Seq454Object.allhits:
            Seq454Object.allhits.append(no_significant_hit_result)
        if not Seq454Object.unculturedhits:
            Seq454Object.unculturedhits.append(no_significant_hit_result)
            
        queue.put(Seq454Object), order.put(i)
        queue.task_done(), order.task_done()
    
    except:
        e = sys.exc_info()[0]
        
        #For now just keeps going past other Errors (usually some issue with connection)
        #Not optimal, but this prevents any issue with saving
        errorlog_filename = (os.path.dirname(os.path.realpath(__file__)) + "\Error Log.txt")     
        errorlog_file = open(errorlog_filename, 'a')
        errorlog_file.write(str(datetime.datetime.now()) + "\n")
        errorlog_file.writelines(str(e))
        errorlog_file.close()
        time.sleep(5)   
        
        #Make a note inside the Blast Results in the 454 Object if there was an error, since
        #errors inside threads will kill the program otherwise, and the desired behavior is to
        #skip the sequence and just go to the next block of sequences. The code will check at a 
        #higher level for this error result in the output.
        
        error_hit_result = BLASTResult(input_query_length=1,
                                       asc = "",
                                       name = "Error Occurred",
                                       score = 0,
                                       evalue = 1e-0,
                                       identities = 0,
                                       subject_as_string = "")
                
        #Fill in any empty lists of hits with no significant hit found message
        if not Seq454Object.tophits:
            Seq454Object.tophits.append(error_hit_result)
        if not Seq454Object.allhits:
            Seq454Object.allhits.append(error_hit_result)
        if not Seq454Object.unculturedhits:
            Seq454Object.unculturedhits.append(error_hit_result)        
        
        queue.put(Seq454Object), order.put(i)
        queue.task_done(), order.task_done()        
        
def OutputResults(list_analyzed_454, project_name, fna_filepaths, boolIncludeAllHits=False, boolIncludedUnculturedHits=False):
    '''Outputs the results of the 454 analysis to .sqlite file'''        
    
    database_path = os.path.dirname(fna_filepaths[0]) + "\\" + project_name + "_Analysis.sqlite"
    con = dbapi.connect(database_path)
    cur = con.cursor()
    
    #Set database output into python to be in regular string format
    con.text_factory = str
    
    #Create Table
    cur.execute('''CREATE TABLE IF NOT EXISTS FirstTopHit
    (SeqID INTEGER NOT NULL, SeqName VARCHAR NOT NULL, ASC VARCHAR,
    MatchName VARCHAR, MatchScore REAL, Identities REAL, Coverage REAL,
    QueryStart INTEGER, QueryEnd INTEGER, SubjectStart INTEGER, SubjectEnd INTEGER,
    PRIMARY KEY (SeqID))''')
    con.commit()
    
    #Populate Table - NOTE: this only uses the top hit if multiple hits are found
    for seq454 in list_analyzed_454:
        cur.execute('''INSERT INTO FirstTopHit VALUES(?,?,?,?,?,?,?,?,?,?,?)''',
                    (None, seq454.name, seq454.tophits[0].asc,
                     seq454.tophits[0].name, seq454.tophits[0].score,
                     seq454.tophits[0].identities, seq454.tophits[0].coverage,
                     seq454.tophits[0].query_start, seq454.tophits[0].query_end,
                     seq454.tophits[0].subject_start, seq454.tophits[0].subject_end))
        
    con.commit()
    
    #Create Table
    cur.execute('''CREATE TABLE IF NOT EXISTS AllTopHits
    (HitID INTEGER NOT NULL, SeqName VARCHAR NOT NULL, ASC VARCHAR, MatchName VARCHAR,
    SequenceMatchNumber INTEGER, MatchScore REAL, Identities REAL, Coverage REAL,
    QueryStart INTEGER, QueryEnd INTEGER, SubjectStart INTEGER, SubjectEnd INTEGER,
    PRIMARY KEY (HitID))''')
    con.commit()
    
    #Populate Table - NOTE: this includes all hits from the "tophits"
    for seq454 in list_analyzed_454:
        hit_number = 0
        for hit in seq454.tophits:
            hit_number += 1
            cur.execute('''INSERT INTO AllTopHits VALUES(?,?,?,?,?,?,?,?,?,?,?,?)''',
                        (None, seq454.name, hit.asc, hit.name, hit_number, hit.score,
                         hit.identities, hit.coverage, hit.query_start,
                         hit.query_end, hit.subject_start, hit.subject_end))
        
    con.commit()    
    
    if boolIncludeAllHits:
        #Create Table
        cur.execute('''CREATE TABLE IF NOT EXISTS AllSignificantHits
        (HitID INTEGER NOT NULL, SeqName VARCHAR NOT NULL, ASC VARCHAR, MatchName VARCHAR,
        SequenceMatchNumber INTEGER, MatchScore REAL, Identities REAL, Coverage REAL,
        QueryStart INTEGER, QueryEnd INTEGER, SubjectStart INTEGER, SubjectEnd INTEGER,
        PRIMARY KEY (HitID))''')
        con.commit()
        
        #Populate Table - NOTE: this includes all hits above minimum values
        for seq454 in list_analyzed_454:
            hit_number = 0
            for hit in seq454.allhits:
                hit_number += 1
                cur.execute('''INSERT INTO AllSignificantHits VALUES(?,?,?,?,?,?,?,?,?,?,?,?)''',
                            (None, seq454.name, hit.asc, hit.name, hit_number, hit.score,
                             hit.identities, hit.coverage, hit.query_start,
                             hit.query_end, hit.subject_start, hit.subject_end))
            
    if boolIncludedUnculturedHits:        
        #Create Table
        cur.execute('''CREATE TABLE IF NOT EXISTS AllUnculturedHits
        (HitID INTEGER NOT NULL, SeqName VARCHAR NOT NULL, ASC VARCHAR, MatchName VARCHAR,
        SequenceMatchNumber INTEGER, MatchScore REAL, Identities REAL, Coverage REAL,
        QueryStart INTEGER, QueryEnd INTEGER, SubjectStart INTEGER, SubjectEnd INTEGER,
        PRIMARY KEY (HitID))''')
        con.commit()
        
        #Populate Table - NOTE: this includes all hits above minimum values
        for seq454 in list_analyzed_454:
            hit_number = 0
            for hit in seq454.unculturedhits:
                hit_number += 1
                cur.execute('''INSERT INTO AllUnculturedHits VALUES(?,?,?,?,?,?,?,?,?,?,?,?)''',
                            (None, seq454.name, hit.asc, hit.name, hit_number, hit.score,
                             hit.identities, hit.coverage, hit.query_start,
                             hit.query_end, hit.subject_start, hit.subject_end))   
        
    con.commit()


    
def FindSkippedSequences(options, window, sql_filepath):
    '''Gets a list of all SequenceData_All table records, and compares to the list of records
    in the FirstTopHit record. Creates the SequenceData table from the ones that aren't in FirstTopHit.
    If FirstTopHit hasn't been created yet, just copies the SequenceData_all table to SequenceData.'''

    database_path = sql_filepath
    con = dbapi.connect(database_path)
    cur = con.cursor()
    con.text_factory = str 
    
    #Check to make sure SequenceData and FirstTopHit tables are in database
    cur.execute('''SELECT tbl_name FROM sqlite_master WHERE type="table";''')
    temp_tablenames = cur.fetchall()
    
    set_tablenames = set()
    for table in temp_tablenames:
        set_tablenames.add(table[0])  
        
    if not ("SequenceData_All" in set_tablenames) and ("SequenceData" in set_tablenames):
        #Adds compatibility for files made before version 1.0.0, by converting sequenceData to SequenceData_All
        cur.execute('''ALTER TABLE SequenceData RENAME TO SequenceData_All''')
        con.commit()        
        
    if ("SequenceData_All" in set_tablenames) and ("FirstTopHit" in set_tablenames):
        #Get SequenceData list of DISTINCT SeqNames
        
        #Get List of Finished Sequences
        cur.execute('''DROP TABLE IF EXISTS FinishedSequences''')
        con.commit()
        cur.execute('''CREATE TABLE IF NOT EXISTS FinishedSequences AS
                       SELECT SeqName FROM FirstTopHit;''')
        con.commit()        
        
        #Recreate the SequenceData table and repopulate with only missing sequences
        cur.execute('''DROP TABLE IF EXISTS SequenceData''')
        con.commit()        
        cur.execute('''CREATE TABLE IF NOT EXISTS SequenceData
                    (SeqID INTEGER NOT NULL, SeqName VARCHAR, Copies INTEGER,
                    Tag VARCHAR, Sequence VARCHAR, CompleteSequence VARCHAR,
                    PRIMARY KEY (SeqID))''')
        con.commit()
        cur.execute('''INSERT INTO SequenceData(SeqID, SeqName, Copies, Tag, Sequence, CompleteSequence)
                    SELECT NULL, SeqName, Copies, Tag, Sequence, CompleteSequence FROM SequenceData_All
                    WHERE SeqName NOT IN FinishedSequences''')
        con.commit()
        
        #Clean up temp tables
        cur.execute('''DROP TABLE IF EXISTS FinishedSequences''')
        con.commit()
        
        cur.execute('''SELECT Count(*) FROM SequenceData''')
        myquery_sequencedata = cur.fetchall() 
        count_tuple = myquery_sequencedata[0]
    
        con.close()
        return count_tuple[0]
    elif ("SequenceData_All" in set_tablenames):
        #Copy SequenceData_All to SequenceData
        cur.execute('''DROP TABLE IF EXISTS SequenceData''')
        con.commit()        
        cur.execute('''CREATE TABLE IF NOT EXISTS SequenceData
                    (SeqID INTEGER NOT NULL, SeqName VARCHAR, Copies INTEGER,
                    Tag VARCHAR, Sequence VARCHAR, CompleteSequence VARCHAR,
                    PRIMARY KEY (SeqID))''')
        con.commit()
        cur.execute('''INSERT INTO SequenceData(SeqID, SeqName, Copies, Tag, Sequence, CompleteSequence)
                    SELECT NULL, SeqName, Copies, Tag, Sequence, CompleteSequence FROM SequenceData_All''')
        con.commit()
        
        #Clean up temp tables
        cur.execute('''DROP TABLE IF EXISTS FinishedSequences''')
        con.commit()        
        
        cur.execute('''SELECT Count(*) FROM SequenceData''')
        myquery_sequencedata = cur.fetchall() 
        count_tuple = myquery_sequencedata[0]

        con.close()
        return count_tuple[0]
    else:
        wx.MessageBox("SequenceData_All table missing from SQL file","Processing Error")
        con.close()
        return -1
    
    
    
def FindTaxonomy(ASC, email, taxonomy_fields = {"superkingdom":"", "phylum":"", "class":"", "order":"", "family":"", "genus":""}):
    '''This function queries entrez to get the full taxonomy for the species given by ASC. Email is needed to use entrez search.
    Function returns the full taxonomy string'''
    searchID = ASC
    entrez.email = email
    
    handle = entrez.efetch(db="nucleotide", id=searchID, rettype="gb", retmode="xml")
    record = entrez.read(handle)
    handle.close
    
    singlerecord = dict()
    singlerecord = record[0]
    
    #Get the species from the organism field GBSeq_organism as "Genus species"
    species = str(singlerecord.get("GBSeq_organism")).split(" ", 1)[1]
    
    #Get the rest of the KNOWN taxonomy from the NCBI taxonomy page (found by ripping the taxon:
    #value from the entrez record
    source = dict()
    source = singlerecord.get("GBSeq_feature-table")[0]
    
    quals = dict()
    quals = source.get("GBFeature_quals")
    
    taxonID = ""
    for x in range(len(quals)):
        qual_dict = dict()
        qual_dict = quals[x]
        
        qualifiervalue = str(qual_dict.get("GBQualifier_value"))
        
        if qualifiervalue.startswith("taxon:"):
            taxonID = qualifiervalue.split(":")[1]
            
    if not taxonID == "":
        #Access the webpage at "http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=" + taxonID
        taxonomy_webpage = "http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=" + taxonID
        
        page = urllib.urlopen(taxonomy_webpage)
        pagetext = page.readlines()
        page.close()
        
        for line in pagetext:
            if "superkingdom" in line:
                if "<STRONG>Lineage</STRONG>" in line:
                    #Handles the case when it gives an ambiguious taxonid
                    #Ex: <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=3077&lvl=3&lin=s&keep=1&srchmode=1&unlock"><STRONG>Lineage</STRONG></A> <SMALL>(full)</SMALL>: <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Root&id=1&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="no rank" ALT="no rank">root</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=131567&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="no rank" ALT="no rank">cellular organisms</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=2759&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="superkingdom" ALT="superkingdom">Eukaryota</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=33090&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="kingdom" ALT="kingdom">Viridiplantae</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=3041&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="phylum" ALT="phylum">Chlorophyta</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=75966&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="class" ALT="class">Trebouxiophyceae</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=35460&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="order" ALT="order">Chlorellales</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=35461&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="family" ALT="family">Chlorellaceae</A>; <A HREF="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&id=3071&lvl=3&lin=f&keep=1&srchmode=1&unlock" TITLE="genus" ALT="genus">Chlorella</A><HR SIZE=1>
                
                    taxon_units = []
                    taxon_units = line.split("</A>")
                    
                    for taxon_unit in taxon_units:
                        p = re.compile("""(?P<href>.+)(" TITLE=")(?P<title>.+)(" ALT=")(?P<taxonomylevel>.+)("\>)(?P<levelname>.+)""")
                        m = p.search(taxon_unit)
                        
                        if m:
                            #Check to see if the taxonomylevel from the regex is on that we want to collect
                            if taxonomy_fields.has_key(m.group("taxonomylevel")):
                                taxonomy_fields[m.group("taxonomylevel")] = m.group("levelname")
                    
                    #There will only be one taxon line per file, add species found earlier and break
                    taxonomy_fields["species"] = species
                    break
                else:
                    #Parse the line for the correct taxonomy fields, first splitting into xml units
                    #Ex: <a ALT="superkingdom" href="/Taxonomy/Browser/wwwtax.cgi?mode=Undef&amp;id=2&amp;lvl=3&amp;keep=1&amp;srchmode=1&amp;unlock" TITLE="superkingdom">Bacteria</a>;
                    p_taxon = re.compile("(?P<htmlstart>\<dd\>{1})(?P<taxonunits>.+)(?P<htmlend>\</a\>\</dd\>{1})")
                    m_taxon = p_taxon.search(line)
                    
                    if m_taxon:
                        taxon_units = []
                        taxon_units = m_taxon.group("taxonunits").split("</a>;")
                        
                        for taxon_unit in taxon_units:
                            p = re.compile("""(\<a ALT=")(?P<taxonomylevel>.+)(" href=")(?P<href>.+)(" TITLE=")(?P<title>.+)("\>)(?P<levelname>.+)""")
                            m = p.search(taxon_unit)
                            
                            if m:
                                #Check to see if the taxonomylevel from the regex is on that we want to collect
                                if taxonomy_fields.has_key(m.group("taxonomylevel")):
                                    taxonomy_fields[m.group("taxonomylevel")] = m.group("levelname")
                    
                    #There will only be one taxon line per file, add species found earlier and break
                    taxonomy_fields["species"] = species
                    break         
        
    return taxonomy_fields
   
if __name__ == '__main__':
    app = wx.PySimpleApp(False)
    frame = MainWindow(None, PROGRAM_NAME + " v" + PROGRAM_VERSION)
    app.MainLoop()  